
import subprocess
import os

locale_folder = "locale"
lang = os.listdir(locale_folder)

for l in lang:

    if l == "messages.pot":
        continue

    po_path = os.path.join(locale_folder, l, "LC_MESSAGES", "tauon.po")
    mo_path = os.path.join(locale_folder, l, "LC_MESSAGES", "tauon.mo")

    if os.path.exists(po_path):
        subprocess.run(['msgfmt', '-o', mo_path, po_path])
        print(f"Compiled: {l}")

    else:
        print(f"Missing po file: {po_path}")

print("Done")




# Tauon Music Box - Album chart image generator

# Copyright © 2015-2019, Taiko2k captain(dot)gxj(at)gmail.com

#     This file is part of Tauon Music Box.
#
#     Tauon Music Box is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     Tauon Music Box is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Lesser General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with Tauon Music Box.  If not, see <http://www.gnu.org/licenses/>.


import gi
gi.require_version('Pango', '1.0')
gi.require_version('PangoCairo', '1.0')
import cairo
from gi.repository import Pango
from gi.repository import PangoCairo
import os
from PIL import Image


class TopChart:

    def __init__(self, tauon, album_art_gen):

        self.pctl = tauon.pctl
        self.cache_dir = tauon.cache_directory
        self.user_dir = tauon.user_directory
        self.album_art_gen = album_art_gen

    def generate(self, tracks, bg=(10,10,10), rows=3, columns=3, show_lables=True, font="Monospace 10", tile=False, cascade=False):

        # Main control variables
        border = 29
        text_offset = -7
        size = 170
        spacing = 9

        mode = 1
        if cascade:
            mode = 2

        if tile:
            border = 0
            spacing = 0
            size = 160
            text_offset = 15

        # Determine the final width and height of album grid
        h = round((border * 2) + (size * rows) + (spacing * (rows - 1)))
        w = round((border * 2) + (size * columns) + (spacing * (columns - 1)))

        if mode == 2:

            r1, r2, r3 = cascade[0]
            print(r1 * 2 + r2 * 2 + r3 * 2)
            sets = []
            for q in range(100, 10000):

                a = q / r1
                b = q / r2
                c = q / r3

                if a - int(a) == b - int(b) == c - int(c) == 0:
                    sets.append((int(a), int(b), int(c)))

            if not sets:
                return False

            abc = None
            for s in sets:
                if s[(r1, r2, r3).index(min((r1, r2, r3)))] > 165:
                    abc = s
                    break
            else:
                return False

            d1, d2, d3 = cascade[1]

            w = round(border * 2) + (abc[0] * r1)
            h = round(border * 2) + (abc[0] * d1) + (abc[1] * d2) + (abc[2] * d3)

        ww = w
        i = -1

        positions = []

        # Grid mode
        if mode == 1:
            for r in range(rows):
                for c in range(columns):

                    i += 1

                    # Break if we run out of albums
                    if i > len(tracks) - 1:
                        break

                    # Determine coordinates for current album
                    x = round(border + ((spacing + size) * c))
                    y = round(border + ((spacing + size) * r))

                    positions.append((tracks[i], x, y, size))

                positions.append(False)

        # Cascade mode
        elif mode == 2:

            a, b, c = abc

            size = a
            spacing = 0
            inv_space = 0
            if not tile:
                inv_space = 8

            x = border
            y = border

            for d in range(d1):
                for cl in range(r1):
                    i += 1
                    x = border + (spacing + size) * cl
                    if i > len(tracks) - 1:
                        break
                    positions.append((tracks[i], x + inv_space // 2, y + inv_space // 2, size - inv_space))
                y += spacing + size

            size = b
            if not tile:
                inv_space = 6
            positions.append(False)

            for d in range(d2):

                for cl in range(r2):
                    i += 1
                    x = border + (spacing + size) * cl
                    print(x)
                    if i > len(tracks) - 1:
                        break
                    positions.append((tracks[i], x + inv_space // 2, y + inv_space // 2, size - inv_space))
                y += spacing + size


            size = c
            if not tile:
                inv_space = 4
            positions.append(False)

            for d in range(d3):
                for cl in range(r3):
                    i += 1
                    x = border + (spacing + size) * cl
                    if i > len(tracks) - 1:
                        break
                    positions.append((tracks[i], x + inv_space // 2, y + inv_space // 2, size - inv_space))
                y += spacing + size


        # Parse font
        font_comp = font.split(" ")
        font_size = int(font_comp.pop())

        col_w = 500

        two_col = False
        font_comp = font.split(" ")
        font_size = int(font_comp.pop())
        if len(positions) * (font_size + 4) > h - border * 2:
            two_col = True
            font_size += 1
            col_w = 380

        if show_lables:
            ww += col_w  # Add extra area to final size for text
            if two_col:
                ww += col_w + 20


        # Get lables
        text = ""
        b_text = ""
        f = False
        for item in positions:

            if item is False:

                if not f:
                    text += " \n"  # Insert extra line to form groups for each row
                else:
                    b_text += " \n"

                if two_col:
                    f ^= True
                continue

            track, x, y, size = item

            # Determine the text label line
            artist = track.artist
            if track.album_artist:
                artist = track.album_artist
            line = f"{artist} - {track.album}\n"

            if not f:
                text += line
            else:
                b_text += line


        # Prepare a blank Cairo surface
        surface = cairo.ImageSurface(cairo.FORMAT_RGB24, ww, h)
        context = cairo.Context(surface)

        bg = (bg[0] / 255, bg[1] / 255, bg[2] / 255)  # Convert 8-bit rgb values to decimal
        context.set_source_rgb(bg[0], bg[1], bg[2])
        context.paint()  # Fill in the background colour

        for item in positions:

            if item is False:
                continue

            track, x, y, size = item

            # Export the album art to file object
            try:
                art_file = self.album_art_gen.save_thumb(track, (size, size), None, png=True, zoom=True)
            except:
                continue

            # Skip drawing this album if loading of artwork failed
            if not art_file:
                continue

            # Load image into Cairo and draw
            art = cairo.ImageSurface.create_from_png(art_file)
            context.set_source_surface(art, x, y)
            context.paint()


        if show_lables:

            # Setup font and prepare Pango layout
            options = context.get_font_options()
            options.set_antialias(cairo.ANTIALIAS_GRAY)
            context.set_font_options(options)
            layout = PangoCairo.create_layout(context)
            layout.set_ellipsize(Pango.EllipsizeMode.END)
            layout.set_width((col_w - text_offset - spacing) * 1000)
            # layout.set_height((h - spacing * 2) * 1000)
            #layout.set_spacing(3 * 1000)
            #layout.set_font_description(Pango.FontDescription(font))


            # Here we make sure the font size is small enough to fit
            font = " ".join(font_comp) + " " + str(font_size)
            layout.set_font_description(Pango.FontDescription(font))
            layout.set_text(text, -1)

            try:
                font_size = int(font_size)
                while font_size > 2:
                    th = layout.get_pixel_size()[1]
                    if th < h - border:
                        break
                    font_size -= 1
                    font = " ".join(font_comp) + " " + str(font_size)
                    layout.set_font_description(Pango.FontDescription(font))
                    layout.set_text(text, -1)
            except:
                print("error adjusting font size")

            # All good to render now
            y_text_padding = 3
            if tile:
                y_text_padding += 6
            context.translate(w + text_offset, border + y_text_padding)
            context.set_source_rgb(0.9, 0.9, 0.9)
            PangoCairo.show_layout(context, layout)

        if b_text:

            # Setup font and prepare Pango layout
            options = context.get_font_options()
            options.set_antialias(cairo.ANTIALIAS_GRAY)
            context.set_font_options(options)
            layout = PangoCairo.create_layout(context)
            layout.set_ellipsize(Pango.EllipsizeMode.END)
            layout.set_width((col_w - text_offset - spacing) * 1000)

            layout.set_font_description(Pango.FontDescription(font))
            layout.set_text(b_text, -1)

            # All good to render now
            y_text_padding = 3
            if tile:
                y_text_padding += 6
            context.translate(col_w + 10, 0)
            context.set_source_rgb(0.9, 0.9, 0.9)
            PangoCairo.show_layout(context, layout)


        # Finally export as PNG
        output_path = os.path.join(self.user_dir, "chart.png")
        surface.write_to_png(output_path)

        # Convert to JPEG for convenience
        output_path2 = os.path.join(self.user_dir, "chart.jpg")
        im = Image.open(output_path)
        im.save(output_path2, 'JPEG', quality=92)

        return output_path2




# Copyright © 2020, Taiko2k captain(dot)gxj(at)gmail.com

#     This file is part of Tauon Music Box.
#
#     Tauon Music Box is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     Tauon Music Box is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Lesser General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with Tauon Music Box.  If not, see <http://www.gnu.org/licenses/>.


import os
try:
    import tekore as tk
    tekore_imported = True
except:
    tekore_imported = False

import pickle
import requests
import io
import webbrowser
import subprocess
import time
import json
from t_modules.t_extra import Timer


class SpotCtl:

    def __init__(self, tauon):
        self.tauon = tauon
        self.strings = tauon.strings
        self.start_timer = Timer()
        self.status = 0
        self.spotify = None
        self.loaded_art = ""
        self.playing = False
        self.coasting = False
        self.paused = False
        self.token = None
        self.cred = None
        self.started_once = False
        self.redirect_uri = f"http://localhost:7811/spotredir"
        self.current_imports = {}
        self.spotify_com = False
        self.sender = None
        self.cache_saved_albums = []
        self.scope = "user-read-playback-position streaming user-modify-playback-state user-library-modify user-library-read user-read-currently-playing user-read-playback-state playlist-read-private playlist-modify-private playlist-modify-public"
        self.launching_spotify = False
        self.progress_timer = Timer()
        self.update_timer = Timer()

        self.token_path = os.path.join(self.tauon.user_directory, "spot-token-pkce")
        self.pkce_code = None

    def prep_cred(self):

        rc = tk.RefreshingCredentials
        self.cred = rc(client_id=self.tauon.prefs.spot_client,
                                    redirect_uri=self.redirect_uri)

    def connect(self):
        if not self.tauon.prefs.spotify_token or not self.tauon.prefs.spot_mode:
            return
        if len(self.tauon.prefs.spot_client) != 32:
            return
        if self.cred is None:
            self.prep_cred()
        if self.spotify is None:
            if self.token is None:
                self.load_token()
            if self.token:
                print("Init spotify support")
                self.sender = tk.RetryingSender(retries=3)
                self.spotify = tk.Spotify(self.token, sender=self.sender)

    def paste_code(self, code):
        if self.cred is None:
            self.prep_cred()

        self.token = self.cred.request_pkce_token(code.strip().strip("\n"), self.pkce_code)
        if self.token:
            self.save_token()
            self.tauon.gui.show_message(self.strings.spotify_account_connected, mode="done")

    def save_token(self):

        if self.token:
            f = open(self.token_path, "w")
            f.write(str(self.token.refresh_token))
            f.close()
            self.tauon.prefs.spotify_token = str(self.token.refresh_token)

    def load_token(self):
        if os.path.isfile(self.token_path):
            f = open(self.token_path, "r")
            self.tauon.prefs.spotify_token = f.read().replace("\n", "").strip()
            f.close()

        if self.tauon.prefs.spotify_token:
            try:
                self.token = tk.refresh_pkce_token(self.tauon.prefs.spot_client, self.tauon.prefs.spotify_token)
            except:
                print("ERROR LOADING TOKEN")
                self.tauon.prefs.spotify_token = ""

    def delete_token(self):
        self.tauon.prefs.spotify_token = ""
        self.token = None
        if os.path.exists(self.token_path):
            os.remove(self.token_path)

    def auth(self):
        if not tekore_imported:
            self.tauon.gui.show_message("python-tekore not installed",
                                        "If you installed via AUR, you'll need to install this optional dependency, then restart Tauon.", mode="error")
            return
        if len(self.tauon.prefs.spot_client) != 32:
            self.tauon.gui.show_message("Invalid client ID. See Spotify tab in settings.", mode="error")
            return
        if self.cred is None:
            self.prep_cred()
        url, self.pkce_code = self.cred.pkce_user_authorisation(scope=self.scope)
        webbrowser.open(url, new=2, autoraise=True)

    def control(self, command, param=None):

        try:
            if command == "pause" and (self.playing or self.coasting) and not self.paused:
                self.spotify.playback_pause()
                self.paused = True
                self.start_timer.set()
            if command == "stop" and (self.playing or self.coasting):
                self.paused = False
                self.playing = False
                self.coasting = False
                self.spotify.playback_pause()
                self.start_timer.set()
            if command == "resume" and (self.coasting or self.playing) and self.paused:
                self.spotify.playback_resume()
                self.paused = False
                self.progress_timer.set()
                self.start_timer.set()
            if command == "volume":
                self.spotify.playback_volume(param)
            if command == "seek":
                self.spotify.playback_seek(param)
                self.start_timer.set()
            if command == "next":
                self.spotify.playback_next(param)
                #self.start_timer.set()
            if command == "previous":
                self.spotify.playback_previous(param)
                #self.start_timer.set()

        except Exception as e:
            print(repr(e))
            if "No active device found" in repr(e):
                try:
                    tr = self.tauon.pctl.playing_object()
                    if command == "resume" and tr and tr.file_ext == "SPTY" and tr.url_key:
                        self.tauon.gui.show_message("Resuming Spotify playback")
                        p = self.tauon.pctl.playing_time
                        self.play_target(tr.url_key)
                        time.sleep(0.3)
                        self.spotify.playback_seek(int(p * 1000))
                        self.tauon.gui.message_box = False
                        self.tauon.gui.update += 1
                        return
                except:
                    pass

                self.tauon.gui.show_message("It looks like there are no more active Spotify devices")

    def add_album_to_library(self, url):
        self.connect()
        if not self.spotify:
            return None

        id = url.strip("/").split("/")[-1]

        try:
            self.spotify.saved_albums_add([id])
            if url not in self.cache_saved_albums:
                self.cache_saved_albums.append(url)
        except:
            print("Error saving album")

    def remove_album_from_library(self, url):

        self.connect()
        if not self.spotify:
            return None
        id = url.strip("/").split("/")[-1]

        try:
            self.spotify.saved_albums_delete([id])
            if url in self.cache_saved_albums:
                self.cache_saved_albums.remove(url)
        except:
            print("Error removing album")

    def get_album_url_from_local(self, track_object):

        if "spotify-album-url" in track_object.misc:
            return track_object.misc["spotify-album-url"]

        self.connect()
        if not self.spotify:
            return None

        results = self.spotify.search(track_object.artist + " " + track_object.album, types=('album',), limit=1)
        for album in results[0].items:
            if "spotify" in album.external_urls:
                return album.external_urls["spotify"]

        return None

    def import_all_playlists(self):

        self.spotify_com = True

        playlists = self.get_playlist_list()
        if playlists:
            for item in playlists:
                self.playlist(item[1], silent=True)
                self.tauon.gui.update += 1
                time.sleep(0.1)

        self.spotify_com = False
        if not playlists:
            self.tauon.gui.show_message(self.strings.spotify_need_enable)
            return
        self.tauon.gui.show_message(self.strings.spotify_import_complete, mode="done")

    def get_playlist_list(self):
        self.connect()
        if not self.spotify:
            self.tauon.gui.show_message(self.strings.spotify_need_enable)
            return None

        playlists = []
        results = self.spotify.playlists(self.spotify.current_user().id)
        pages = self.spotify.all_pages(results)
        for page in pages:
            items = page.items
            for item in items:
                name = item.name
                url = item.external_urls["spotify"]
                playlists.append((name, url))

        return playlists

    def search(self, text):
        self.connect()
        if not self.spotify:
            return

        results = self.spotify.search(text,
                                      types=('artist', 'album', 'track'),
                                      limit=20
                                      )

        finds = []

        self.tauon.QuickThumbnail.queue.clear()

        if results[0]:

            for i, album in enumerate(results[0].items[1:]):

                img = self.tauon.QuickThumbnail()
                img.url = album.images[-1].url
                img.size = round(50 * self.tauon.gui.scale)
                self.tauon.QuickThumbnail().items.append(img)
                if i < 10:
                    self.tauon.QuickThumbnail().queue.append(img)
                try:
                    self.tauon.gall_ren.lock.release()
                except:
                    pass

                finds.append((11, (album.name, album.artists[0].name), album.external_urls["spotify"], 0, 0, img))

            for artist in results[1].items[0:1]:
                finds.insert(1, (10, artist.name, artist.external_urls["spotify"], 0, 0, None))
            for artist in results[1].items[1:2]:
                finds.insert(11, (10, artist.name, artist.external_urls["spotify"], 0, 0, None))

            for track in results[2].items[0:1]:
                finds.insert(2, (12, (track.name, track.artists[0].name), track.external_urls["spotify"], 0, 0, None))
            for track in results[2].items[5:6]:
                finds.insert(8, (12, (track.name, track.artists[0].name), track.external_urls["spotify"], 0, 0, None))

        return finds


    def search_track(self, track):
        if track is None:
            return

        self.connect()
        if not self.spotify:
            return

        if track.artist and track.title:
            results = self.spotify.search(track.artist + " " + track.title,
                                     types=('track',),
                                     limit=1
                                     )

    def prime_device(self):
        self.connect()
        if not self.spotify:
            return

        devices = self.spotify.playback_devices()

        if devices:
            pass
        else:
            print("No spotify devices found")

        if not devices:
            return False
        for d in devices:
            if d.is_active:
                return None
        for d in devices:
            if not d.is_restricted:
                return d.id
        return None

    def play_target(self, id):

        self.coasting = False
        self.connect()
        if not self.spotify:
            self.tauon.gui.show_message("Error. You may need to click Authorise in Settings > Accounts > Spotify.", mode="warning")
            return

        d_id = self.prime_device()
        # if d_id is False:
        #     return


        #if self.tauon.pctl.playing_state == 1 and self.playing and self.tauon.pctl.playing_time
        #try:
        if d_id is False:
            self.launching_spotify = True
            self.tauon.gui.update += 1
            if self.tauon.prefs.launch_spotify_web:
                webbrowser.open("https://open.spotify.com/", new=2, autoraise=False)
                tries = 0
                while True:
                    time.sleep(2)
                    if tries == 0:
                        self.tauon.focus_window()
                    devices = self.spotify.playback_devices()
                    if devices:
                        self.progress_timer.set()
                        self.spotify.playback_start_tracks([id], device_id=devices[0].id)
                        break
                    tries += 1
                    if tries > 6:
                        self.tauon.pctl.stop()
                        self.tauon.gui.show_message(self.strings.spotify_error_starting, mode="error")
                        self.launching_spotify = False
                        self.tauon.gui.update += 1
                        return
            else:
                subprocess.run(["xdg-open", "spotify:track"])
                print("LAUNCH SPOTIFY")
                time.sleep(3)
                tries = 0
                playing = False
                while True:
                    print("WAIT FOR DEVICE...")
                    devices = self.spotify.playback_devices()
                    if devices:
                        print("DEVICE FOUND")
                        self.tauon.focus_window()
                        time.sleep(1)
                        print("ATTEMPT START")

                        self.spotify.playback_start_tracks([id], device_id=devices[0].id)
                        while True:
                            result = self.spotify.playback_currently_playing()
                            if result and result.is_playing:
                                playing = True
                                self.progress_timer.set()
                                print("TRACK START SUCCESS")
                                break
                            time.sleep(2)
                            tries += 1
                            print("NOT PLAYING YET...")
                            if tries > 6:
                                break
                    if playing:
                        break
                    tries += 1
                    if tries > 6:
                        print("TOO MANY TRIES")
                        self.tauon.pctl.stop()
                        self.tauon.gui.show_message(self.strings.spotify_error_starting, mode="error")
                        self.launching_spotify = False
                        self.tauon.gui.update += 1
                        return
                    time.sleep(2)

            self.launching_spotify = False
            self.tauon.gui.update += 1

        else:
            try:
                self.progress_timer.set()
                okay = False

                # Check conditions for a proper transition
                if self.playing:
                    result = self.spotify.playback_currently_playing()
                    if result and result.item and result.is_playing:
                        remain = result.item.duration_ms - result.progress_ms
                        if 1400 < remain < 3500:
                            self.spotify.playback_queue_add("spotify:track:" + id,  device_id=d_id)
                            okay = True
                            time.sleep(remain / 1000)
                            self.progress_timer.set()
                            time.sleep(1)
                            result = self.spotify.playback_currently_playing()
                            if not (result and result.item and result.is_playing):
                                print("A queue transition failed")
                                okay = False

                # Force a transition
                if not okay:
                    self.spotify.playback_start_tracks([id], device_id=d_id)

            # except tk.client.decor.error.InternalServerError:
            #     self.tauon.gui.show_message("Spotify server error. Maybe try again later.")
            #     return
            except:
                self.tauon.gui.show_message("Spotify error, try again?", mode="warning")
                return
        # except Exception as e:
        #     self.tauon.gui.show_message("Error. Do you have playback started somewhere?", mode="error")
        self.playing = True
        self.started_once = True

        self.start_timer.set()
        self.tauon.pctl.playing_time = 0
        self.tauon.pctl.decode_time = 0
        self.tauon.gui.pl_update += 1
        self.tauon.tm.ready_playback()

    def get_library_albums(self, return_list=False):
        self.connect()
        if not self.spotify:
            self.spotify_com = False
            self.tauon.gui.show_message(self.strings.spotify_need_enable)
            return []

        albums = self.spotify.saved_albums()

        playlist = []
        self.update_existing_import_list()
        self.cache_saved_albums.clear()

        pages = self.spotify.all_pages(albums)

        for page in pages:
            for a in page.items:
                self.load_album(a.album, playlist)

                if a.album.external_urls["spotify"] not in self.cache_saved_albums:
                    self.cache_saved_albums.append(a.album.external_urls["spotify"])

        if return_list:
            return playlist

        self.tauon.pctl.multi_playlist.append(self.tauon.pl_gen(title=self.strings.spotify_albums, playlist=playlist))
        self.tauon.pctl.gen_codes[self.tauon.pl_to_id(len(self.tauon.pctl.multi_playlist) - 1)] = "sal"
        self.spotify_com = False

    def append_track(self, url):

        self.connect()
        if not self.spotify:
            return

        if url.startswith("spotify:track:"):
            id = url[14:]
        else:
            url = url.split("?")[0]
            id = url.strip("/").split("/")[-1]

        track = self.spotify.track(id)
        tr = self.load_track(track)
        self.tauon.pctl.master_library[tr.index] = tr
        self.tauon.pctl.multi_playlist[self.tauon.pctl.active_playlist_viewing][2].append(tr.index)
        self.tauon.gui.pl_update += 1

    def append_album(self, url, playlist_number=None, return_list=False):

        self.connect()
        if not self.spotify:
            return

        if url.startswith("spotify:album:"):
            id = url[14:]
        else:
            url = url.split("?")[0]
            id = url.strip("/").split("/")[-1]

        album = self.spotify.album(id)
        playlist = []
        self.update_existing_import_list()
        self.load_album(album, playlist)

        if return_list:
            return playlist

        if playlist_number is None:
            playlist_number = self.tauon.pctl.active_playlist_viewing

        self.tauon.pctl.multi_playlist[playlist_number][2].extend(playlist)
        self.tauon.gui.pl_update += 1

    def playlist(self, url, return_list=False, silent=False):

        self.connect()
        if not self.spotify:
            return []

        if url.startswith("spotify:playlist:"):
            id = url[17:]
        else:
            url = url.split("?")[0]
            if len(url) != 22:
                id = url.strip("/").split("/")[-1]
            else:
                id = url

        if len(id) != 22:
            print("ID Error")
            if return_list:
                return []
            return

        p = self.spotify.playlist(id)
        playlist = []
        self.update_existing_import_list()
        pages = self.spotify.all_pages(p.tracks)
        for page in pages:
            for item in page.items:
                nt = self.load_track(item.track, include_album_url=True)
                self.tauon.pctl.master_library[nt.index] = nt
                playlist.append(nt.index)

        if return_list:
            return playlist

        title = p.name + " by " + p.owner.display_name
        self.tauon.pctl.multi_playlist.append(self.tauon.pl_gen(title=title, playlist=playlist))
        if p.name == "Discover Weekly" or p.name == "Release Radar":
            self.tauon.pctl.multi_playlist[len(self.tauon.pctl.multi_playlist) - 1][4] = 1

        self.tauon.pctl.gen_codes[self.tauon.pl_to_id(len(self.tauon.pctl.multi_playlist) - 1)] = f"spl\"{id}\""
        if not silent:
            self.tauon.switch_playlist(len(self.tauon.pctl.multi_playlist) - 1)

    def artist_playlist(self, url):
        id = url.strip("/").split("/")[-1]
        artist = self.spotify.artist(id)
        artist_albums = self.spotify.artist_albums(id, limit=50, include_groups=["album"])
        playlist = []
        self.update_existing_import_list()

        for a in artist_albums.items:
            full_album = self.spotify.album(a.id)
            self.load_album(full_album, playlist)

        self.tauon.pctl.multi_playlist.append(self.tauon.pl_gen(title="Spotify: " + artist.name, playlist=playlist))
        self.tauon.switch_playlist(len(self.tauon.pctl.multi_playlist) - 1)
        self.tauon.gui.message_box = False

    def update_existing_import_list(self):
        self.current_imports.clear()
        for tr in self.tauon.pctl.master_library.values():
            if "spotify-track-url" in tr.misc:
                self.current_imports[tr.misc["spotify-track-url"]] = tr

    def create_playlist(self, name):
        print("Create new spotify playlist")
        self.connect()
        if not self.spotify:
            return None

        try:
            user = self.spotify.current_user()
            playlist = self.spotify.playlist_create(user.id, name, True)
            return playlist.id
        except:
            return None

    def upload_playlist(self, playlist_id, track_urls):
        self.connect()
        if not self.spotify:
            return None

        try:
            uris = []
            for url in track_urls:
                uris.append("spotify:track:" + url.strip("/").split("/")[-1])

            self.spotify.playlist_clear(playlist_id)
            time.sleep(0.05)
            with self.spotify.chunked(True):
                self.spotify.playlist_add(playlist_id, uris)
        except:
            self.tauon.gui.show_message("Spotify upload error!", mode="error")

    def load_album(self, album, playlist):
        #a = item
        album_url = album.external_urls["spotify"]
        art_url = album.images[0].url
        album_name = album.name
        total_tracks = album.total_tracks
        date = album.release_date
        album_artist = album.artists[0].name
        id = album.id
        parent = (album_artist + " - " + album_name).strip("- ")

        # print(a.release_date, a.name)
        for track in album.tracks.items:

            pr = None
            if "spotify" in track.external_urls:
                pr = self.current_imports.get(track.external_urls["spotify"])
            if pr:
                new = False
                nt = pr
            else:
                new = True
                nt = self.tauon.TrackClass()
                nt.index = self.tauon.pctl.master_count

            nt.is_network = True
            nt.file_ext = "SPTY"
            nt.url_key = track.id
            if track.artists and "spotify" in track.artists[0].external_urls:
                nt.misc["spotify-artist-url"] = track.artists[0].external_urls["spotify"]
            nt.misc["spotify-album-url"] = album_url
            if "spotify" in track.external_urls:
                nt.misc["spotify-track-url"] = track.external_urls["spotify"]
            nt.artist = track.artists[0].name
            nt.album_artist = album_artist
            nt.date = date
            nt.album = album_name
            nt.disc_number = track.disc_number
            #nt.disc_total =
            nt.length = track.duration_ms / 1000
            nt.title = track.name
            nt.track_number = track.track_number
            nt.track_total = total_tracks
            nt.art_url_key = art_url
            nt.parent_folder_path = parent
            nt.parent_folder_name = parent
            if new:
                self.tauon.pctl.master_count += 1
                self.tauon.pctl.master_library[nt.index] = nt
            playlist.append(nt.index)



    def load_track(self, track, update_master_count=True, include_album_url=False):
        if "spotify" in track.external_urls:
            pr = self.current_imports.get(track.external_urls["spotify"])
        
        else:
            pr = False

        if pr:
            new = False
            nt = pr
        else:
            new = True
            nt = self.tauon.TrackClass()
            nt.index = self.tauon.pctl.master_count

        nt.is_network = True
        nt.file_ext = "SPTY"
        nt.url_key = track.id
        #if new:
        if "spotify" in track.artists[0].external_urls:
            nt.misc["spotify-artist-url"] = track.artists[0].external_urls["spotify"]
        if include_album_url and "spotify-album-url" not in nt.misc:
            if "spotify" in track.album.external_urls:
                nt.misc["spotify-album-url"] = track.album.external_urls["spotify"]
        if "spotify" in track.external_urls:
            nt.misc["spotify-track-url"] = track.external_urls["spotify"]
        if track.artists[0].name:
            nt.artist = track.artists[0].name
        if track.album.artists:
            nt.album_artist = track.album.artists[0].name
        if track.album.release_date:
            nt.date = track.album.release_date
        nt.album = track.album.name
        nt.disc_number = track.disc_number
        nt.length = track.duration_ms / 1000
        nt.title = track.name
        nt.track_number = track.track_number
        # nt.track_total = total_tracks
        if track.album.images:
            nt.art_url_key = track.album.images[0].url
        parent = (nt.album_artist + " - " + nt.album).strip("- ")
        nt.parent_folder_path = parent
        nt.parent_folder_name = parent

        if update_master_count and new:
            self.tauon.pctl.master_count += 1

        return nt

    def like_track(self, tract_object):
        track_url = tract_object.misc.get("spotify-track-url", False)
        if track_url:
            id = track_url.strip("/").split("/")[-1]
            results = self.spotify.saved_tracks_contains([id])
            if not results or results[0] is False:
                self.spotify.saved_tracks_add([id])
                tract_object.misc["spotify-liked"] = True
                self.tauon.gui.show_message(self.strings.spotify_like_added, mode="done")
                return
            self.tauon.gui.show_message(self.strings.spotify_already_liked)
            return

    def unlike_track(self, tract_object):
        track_url = tract_object.misc.get("spotify-track-url", False)
        if track_url:
            id = track_url.strip("/").split("/")[-1]
            results = self.spotify.saved_tracks_contains([id])
            if not results or results[0] is True:
                self.spotify.saved_tracks_delete([id])
                tract_object.misc.pop("spotify-liked", None)
                self.tauon.gui.show_message(self.strings.spotify_un_liked, mode="done")
                return
            self.tauon.gui.show_message(self.strings.spotify_already_un_liked)
            return

    def get_library_likes(self, return_list=False):
        self.connect()
        if not self.spotify:
            self.spotify_com = False
            self.tauon.gui.show_message(self.strings.spotify_need_enable)
            return []

        self.update_existing_import_list()
        tracks = self.spotify.saved_tracks()

        playlist = []

        for tr in self.tauon.pctl.master_library.values():
            tr.misc.pop("spotify-liked", None)

        pages = self.spotify.all_pages(tracks)
        for page in pages:
            for item in page.items:
                nt = self.load_track(item.track)
                self.tauon.pctl.master_library[nt.index] = nt
                playlist.append(nt.index)
                nt.misc["spotify-liked"] = True

        if return_list:
            return playlist

        for p in self.tauon.pctl.multi_playlist:
            if p[0] == self.tauon.strings.spotify_likes:
                p[2][:] = playlist[:]
                self.spotify_com = False
                return

        self.tauon.pctl.multi_playlist.append(self.tauon.pl_gen(title=self.tauon.strings.spotify_likes, playlist=playlist))
        self.tauon.pctl.gen_codes[self.tauon.pl_to_id(len(self.tauon.pctl.multi_playlist) - 1)] = "slt"
        self.spotify_com = False

    def monitor(self):
        tr = self.tauon.pctl.playing_object()
        result = None

        # Detect if playback has resumed
        if self.playing and self.paused:
            result = self.spotify.playback_currently_playing()
            if result and result.is_playing:
                self.paused = False
                self.progress_timer.set()
                self.tauon.pctl.playing_state = 1
                self.tauon.gui.update += 1

        # Detect is playback has been modified
        elif self.playing and self.start_timer.get() > 4 and self.tauon.pctl.playing_time + 5 < tr.length:

            if not result:
                result = self.spotify.playback_currently_playing()

            # Playback has been stopped?
            if (result is None or result.item is None) or tr is None:
                self.playing = False
                self.tauon.pctl.stop()
                return
            # Playback has been paused?
            elif tr and result and not result.is_playing:
                self.paused = True
                self.tauon.pctl.playing_state = 2
                self.tauon.gui.update += 1
                return
            # Something else is now playing? If so, switch into coast mode
            if result.item.name != tr.title:
                self.tauon.pctl.playing_state = 3
                self.playing = False
                self.coasting = True
                self.coast_update(result)
                self.tauon.gui.pl_update += 2
                return

            p = result.progress_ms
            if p is not None:
                #if abs(self.tauon.pctl.playing_time - (p / 1000)) > 0.4:
                    # print("DESYNC")
                    # print(abs(self.tauon.pctl.playing_time - (p / 1000)))
                self.tauon.pctl.playing_time = p / 1000
                self.tauon.pctl.decode_time = self.tauon.pctl.playing_time
                # else:
                #     print("SYNCED")

    def update(self, start=False):

        if self.playing:
            self.coasting = False
            return

        self.connect()
        if not self.spotify:
            return

        result = self.spotify.playback_currently_playing()
        self.tauon.tm.ready_playback()

        if self.playing or (not self.coasting and not start):
            return

        try:
            self.tauon.tm.player_lock.release()
        except:
            pass

        if result is None or result.is_playing is False:
            if self.coasting:

                if self.tauon.pctl.radio_image_bin:
                    self.loaded_art = ""
                    self.tauon.pctl.radio_image_bin.close()
                    self.tauon.pctl.radio_image_bin = None
                    self.tauon.dummy_track.artist = ""
                    self.tauon.dummy_track.date = ""
                    self.tauon.dummy_track.title = ""
                    self.tauon.dummy_track.album = ""
                    self.tauon.dummy_track.art_url_key = ""
                    self.tauon.gui.clear_image_cache_next = True
                    self.paused = True

            else:
                self.tauon.gui.show_message(self.strings.spotify_not_playing)
            return

        self.coasting = True
        self.started_once = True
        self.tauon.pctl.playing_state = 3

        if result.is_playing:
            self.paused = False
        else:
            self.paused = True

        self.coast_update(result)

    def append_playing(self, playlist_number):
        if not self.coasting:
            return
        tr = self.tauon.pctl.playing_object()
        if tr and "spotify-album-url" in tr.misc:
            self.append_album(tr.misc["spotify-album-url"], playlist_number)

    def coast_update(self, result):

        if result is None or result.item is None:
            print("Spotify returned unknown")
            return

        self.tauon.dummy_track.artist = result.item.artists[0].name
        self.tauon.dummy_track.title = result.item.name
        self.tauon.dummy_track.album = result.item.album.name
        self.tauon.dummy_track.date = result.item.album.release_date
        self.tauon.dummy_track.file_ext = "Spotify"

        self.progress_timer.set()
        self.update_timer.set()

        d = result.item.duration_ms
        if d is not None:
            self.tauon.pctl.playing_length = d / 1000

        p = result.progress_ms
        if p is not None:
            self.tauon.pctl.playing_time = p / 1000

        self.tauon.pctl.decode_time = self.tauon.pctl.playing_time

        art_url = result.item.album.images[0].url
        self.tauon.dummy_track.url_key = result.item.id
        self.tauon.dummy_track.misc["spotify-album-url"] = result.item.album.external_urls["spotify"]
        self.tauon.dummy_track.misc["spotify-track-url"] = result.item.external_urls["spotify"]

        if art_url and self.loaded_art != art_url:
            self.loaded_art = art_url
            art_response = requests.get(art_url)
            if self.tauon.pctl.radio_image_bin:
                self.tauon.pctl.radio_image_bin.close()
                self.tauon.pctl.radio_image_bin = None
            self.tauon.pctl.radio_image_bin = io.BytesIO(art_response.content)
            self.tauon.pctl.radio_image_bin.seek(0)
            self.tauon.dummy_track.art_url_key = "ok"
            self.tauon.gui.clear_image_cache_next = True

        self.tauon.gui.update += 2
        self.tauon.gui.pl_update += 1





import subprocess
import os

locale_folder = "locale"
pot_path = os.path.join(locale_folder, "messages.pot")

print("Generate template")
subprocess.run(['python', "pygettext.py", "t_modules/t_main.py"])
print("Copy template")
subprocess.run(['cp', "messages.pot", pot_path])
subprocess.run(['rm', "messages.pot"])

lang = os.listdir(locale_folder)

for l in lang:

    if l == "messages.pot":
        continue

    po_path = os.path.join(locale_folder, l, "LC_MESSAGES", "tauon.po")
    
    if os.path.exists(po_path):
        subprocess.run(['msgmerge', '-U', po_path, pot_path])

        print(f"Updated: {l}")

    else:
        print(f"Missing po file: {po_path}")

print("Done")





# Tauon Music Box - Basic Drawing and Text Drawing Functions Module

# Copyright © 2015-2019, Taiko2k captain(dot)gxj(at)gmail.com

#     This file is part of Tauon Music Box.
#
#     Tauon Music Box is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     Tauon Music Box is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Lesser General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with Tauon Music Box.  If not, see <http://www.gnu.org/licenses/>.
import math
import sys
from sdl2 import *
from sdl2.sdlimage import IMG_Load_RW
from t_modules.t_extra import *
import ctypes
from ctypes import pointer
import io
from PIL import Image

system = "linux"
if sys.platform == 'win32':
    system = "windows"
system = "linux"
if system == "linux":
    import cairo
    import gi
    gi.require_version('Pango', '1.0')
    gi.require_version('PangoCairo', '1.0')
    from gi.repository import Pango
    from gi.repository import PangoCairo
else:
    from ctypes import windll, CFUNCTYPE, POINTER, c_int, c_void_p, byref, pointer
    import win32con, win32api, win32gui, win32ui
    import struct


class QuickThumbnail:

    renderer = None
    items = []
    queue = []

    def __init__(self):
        self.rect = SDL_Rect(0, 0)
        self.texture = None
        self.surface = None
        self.size = 50
        self.alive = False
        self.url = None

    def destruct(self):
        if self.surface:
            SDL_FreeSurface(self.surface)
            self.surface = None
        if self.texture:
            SDL_DestroyTexture(self.texture)
            self.texture = None
        self.alive = False

    def read_and_thumbnail(self, f, w, h):

        g = io.BytesIO()
        g.seek(0)
        im = Image.open(f)
        im.thumbnail((w, h), Image.ANTIALIAS)
        im.save(g, 'PNG')
        g.seek(0)
        wop = rw_from_object(g)
        self.surface = IMG_Load_RW(wop, 0)
        #self.items.append(self)
        self.alive = True

    def prime(self):

        texture = SDL_CreateTextureFromSurface(self.renderer, self.surface)
        SDL_FreeSurface(self.surface)
        self.surface = None
        tex_w = pointer(c_int(0))
        tex_h = pointer(c_int(0))
        SDL_QueryTexture(texture, None, None, tex_w, tex_h)
        self.rect.w = int(tex_w.contents.value)
        self.rect.h = int(tex_h.contents.value)
        self.texture = texture

    def draw(self, x, y):
        if len(self.items) > 30:
            img = self.items[0]
            img.destruct()
            self.items.remove(img)
        if not self.alive:
            if self not in self.queue:
                self.queue.append(self)
            return False
        if not self.texture:
            self.prime()
        self.rect.x = round(x)
        self.rect.y = round(y)
        SDL_RenderCopy(self.renderer, self.texture, None, self.rect)

        return True

if system == "windows":

    class RECT(ctypes.Structure):
        _fields_ = [('left', ctypes.c_long),
                    ('top', ctypes.c_long),
                    ('right', ctypes.c_long),
                    ('bottom', ctypes.c_long)
                    ]

    def RGB(r, g, b):
        return r | (g << 8) | (b << 16)

    def Wcolour(colour):
        return colour[0] | (colour[1] << 8) | (colour[2] << 16)

    def native_bmp_to_sdl(hdc, bitmap_handle, width, height):

        bmpheader = struct.pack("LHHHH", struct.calcsize("LHHHH"),
                                width, height, 1, 24) #w,h, planes=1, bitcount)

        c_bmpheader = ctypes.c_buffer(bmpheader)

        #3 bytes per pixel, pad lines to 4 bytes
        c_bits = ctypes.c_buffer(b" " * (height * ((width*3 + 3) & -4)))

        res = ctypes.windll.gdi32.GetDIBits(
            hdc, bitmap_handle, 0, height,
            c_bits, c_bmpheader,
            win32con.DIB_RGB_COLORS)

        if not res:
            raise IOError("native_bmp_to_pil failed: GetDIBits")

        # We need to keep c_bits pass else it may be garbage collected
        return SDL_CreateRGBSurfaceWithFormatFrom(ctypes.pointer(c_bits), width, height, 24, (width*3 + 3) & -4 , SDL_PIXELFORMAT_BGR24), c_bits


    class Win32Font:

        def __init__(self, name, height, weight=win32con.FW_NORMAL,
                     italic=False, underline=False):
                     
            self.font = win32ui.CreateFont({
                'name': name, 'height': height,
                'weight': weight, 'italic': italic, 'underline': underline,}) #'charset': win32con.MAC_CHARSET})

            #create a compatible DC we can use to draw:

            self.desktopHwnd = win32gui.GetDesktopWindow()
            self.desktopDC = win32gui.GetWindowDC(self.desktopHwnd)
            self.mfcDC = win32ui.CreateDCFromHandle(self.desktopDC)
            self.drawDC = self.mfcDC.CreateCompatibleDC()

            #initialize it

            self.drawDC.SelectObject(self.font)

        def get_metrics(self, text, max_x, wrap):

            #return self.drawDC.GetTextExtent(text)
            
            rect = RECT(0,0,0,0)
            rect.left = 0
            rect.right = round(max_x)
            rect.top = 0
            rect.bottom = 0

                #windll.User32.DrawTextW(t, text, len(text)) #, rect, win32con.DT_WORDBREAK)
            t = self.drawDC.GetSafeHdc()
            
            if wrap:
        
                windll.User32.DrawTextW(t, text, len(text), pointer(rect), win32con.DT_WORDBREAK | win32con.DT_CALCRECT)
            else:
                windll.User32.DrawTextW(t, text, len(text), pointer(rect), win32con.DT_CALCRECT | win32con.DT_END_ELLIPSIS)
                
            return rect.right, rect.bottom
            

        def renderText(self, text, bg, fg, wrap=False, max_x=100, max_y=None):

            self.drawDC.SetTextColor(Wcolour(fg))

            t = self.drawDC.GetSafeHdc()

            win32gui.SetBkMode(t, win32con.TRANSPARENT)

            #create the compatible bitmap:

            #w,h = self.drawDC.GetTextExtent(text)
            w, h = self.get_metrics(text, max_x, wrap)

            #print(self.drawDC.GetTextFace())

            #w += 1
            #if wrap:
            #    h = int((w / max_x) * h) + h
            #    w = max_x + 1
            if max_y != None:
                h = max_y

            saveBitMap = win32ui.CreateBitmap()

            saveBitMap.CreateCompatibleBitmap(self.mfcDC, w, h)

            self.drawDC.SelectObject(saveBitMap)

            #draw it

            br = win32ui.CreateBrush(win32con.BS_SOLID, Wcolour(bg), 0)

            self.drawDC.FillRect((0, 0, w, h), br)

            #self.drawDC.DrawText(text, (0, 0, w, h), win32con.DT_LEFT)

            #windll.gdi32.TextOutW(t, 0, 0, "test", 5)

            if wrap:
                rect = RECT(0,0,0,0)
                rect.left = 0
                rect.right = round(max_x)
                rect.top = 0
                rect.bottom = round(h)

                #windll.User32.DrawTextW(t, text, len(text)) #, rect, win32con.DT_WORDBREAK)
                windll.User32.DrawTextW(t, text, len(text), pointer(rect), win32con.DT_WORDBREAK)
            else:
                
                rect = RECT(0,0,0,0)
                rect.left = 0
                rect.right = round(max_x)
                rect.top = 0
                rect.bottom = round(h)

                #windll.User32.DrawTextW(t, text, len(text)) #, rect, win32con.DT_WORDBREAK)
                windll.User32.DrawTextW(t, text, len(text), pointer(rect), win32con.DT_END_ELLIPSIS)                
                
                
                #windll.gdi32.TextOutW(t, 0, 0, text, len(text))

            # print(rects)
            #print(text)
            #windll.gdi32.ExtTextOutW(t, 0, 0, None, rect, text, len(text), None)
            #convert to SDL surface
            im, c_bits = native_bmp_to_sdl(self.drawDC.GetSafeHdc(), saveBitMap.GetHandle(), w, h)
            #clean-up
            win32gui.DeleteObject(saveBitMap.GetHandle())

            return im, c_bits


        def __del__(self):

            self.mfcDC.DeleteDC()
            self.drawDC.DeleteDC()
            win32gui.ReleaseDC(self.desktopHwnd, self.desktopDC)
            win32gui.DeleteObject(self.font.GetSafeHandle())

        def __del__(self):

            win32gui.DeleteObject(self.font.GetSafeHandle())



# Performs alpha blending of one colour (RGB-A) onto another (RGB)
def alpha_blend(colour, base):
    alpha = colour[3] / 255
    return [int(alpha * colour[0] + (1 - alpha) * base[0]),
            int(alpha * colour[1] + (1 - alpha) * base[1]),
            int(alpha * colour[2] + (1 - alpha) * base[2]), 255]


perf = Timer()

class TDraw:

    def __init__(self, renderer=None):

        # All
        self.renderer = renderer
        self.scale = 1
        self.force_subpixel_text = False

        # Drawing
        self.sdl_rect = SDL_Rect(10, 10, 10, 10)

        # Text and Fonts
        self.source_rect = SDL_Rect(0, 0, 0, 0)
        self.dest_rect = SDL_Rect(0, 0, 0, 0)


        if system == "linux":
            self.surf = cairo.ImageSurface(cairo.FORMAT_ARGB32, 0, 0)
            self.context = cairo.Context(self.surf)
            self.layout = PangoCairo.create_layout(self.context)
            
        else:
            self.cache = {}
            self.ca_li = []
            self.y_offset_dict = {}

        self.text_background_colour = [0, 0, 0, 255]
        #self.pretty_rect = [0, 0, 0, 0]
        self.pretty_rect = None
        self.real_bg = False
        self.alpha_bg = False
        self.force_gray = False
        self.f_dict = {}
        self.ttc = {}
        self.ttl = []

        self.was_truncated = False

    def rect_s(self, rectangle, colour, thickness):
        SDL_SetRenderDrawColor(self.renderer, colour[0], colour[1], colour[2], colour[3])
        x, y, w, h = (round(x) for x in rectangle)
        th = math.floor(thickness)
        self.sdl_rect.x = x - th
        self.sdl_rect.y = y - th
        self.sdl_rect.w = th
        self.sdl_rect.h = h + th
        SDL_RenderFillRect(self.renderer, self.sdl_rect) # left
        self.sdl_rect.x = x - th
        self.sdl_rect.y = y + h
        self.sdl_rect.w = w + th
        self.sdl_rect.h = th
        SDL_RenderFillRect(self.renderer, self.sdl_rect) # bottom
        self.sdl_rect.x = x
        self.sdl_rect.y = y - th
        self.sdl_rect.w = w + th
        self.sdl_rect.h = th
        SDL_RenderFillRect(self.renderer, self.sdl_rect) # top
        self.sdl_rect.x = x + w
        self.sdl_rect.y = y
        self.sdl_rect.w = th
        self.sdl_rect.h = h + th
        SDL_RenderFillRect(self.renderer, self.sdl_rect) # right

    def rect_a(self, location_xy, size_wh, colour):
        self.rect((location_xy[0], location_xy[1], size_wh[0], size_wh[1]), colour)

    def rect(self, rectangle, colour):

        self.sdl_rect.x = round(rectangle[0])
        self.sdl_rect.y = round(rectangle[1])
        self.sdl_rect.w = round(rectangle[2])
        self.sdl_rect.h = round(rectangle[3])

        SDL_SetRenderDrawColor(self.renderer, colour[0], colour[1], colour[2], colour[3])

        #if fill:
        SDL_RenderFillRect(self.renderer, self.sdl_rect)
        # else:
        #     SDL_RenderDrawRect(self.renderer, self.sdl_rect)

    def bordered_rect(self, rectangle, fill_colour, outer_colour, border_size):

        self.sdl_rect.x = round(rectangle[0]) - border_size
        self.sdl_rect.y = round(rectangle[1]) - border_size
        self.sdl_rect.w = round(rectangle[2]) + border_size + border_size
        self.sdl_rect.h = round(rectangle[3]) + border_size + border_size
        SDL_SetRenderDrawColor(self.renderer, outer_colour[0], outer_colour[1], outer_colour[2], outer_colour[3])
        SDL_RenderFillRect(self.renderer, self.sdl_rect)
        self.sdl_rect.x = round(rectangle[0])
        self.sdl_rect.y = round(rectangle[1])
        self.sdl_rect.w = round(rectangle[2])
        self.sdl_rect.h = round(rectangle[3])
        SDL_SetRenderDrawColor(self.renderer, fill_colour[0], fill_colour[1], fill_colour[2], fill_colour[3])
        SDL_RenderFillRect(self.renderer, self.sdl_rect)

    def line(self, x1, y1, x2, y2, colour):

        SDL_SetRenderDrawColor(self.renderer, colour[0], colour[1], colour[2], colour[3])
        SDL_RenderDrawLine(self.renderer, round(x1), round(y1), round(x2), round(y2))

    def get_text_w(self, text, font, height=False):

        x, y = self.get_text_wh(text, font, 3000)
        if height:
            return y
        else:
            return x

    def clear_text_cache(self):

        for key in self.ttl:
            so = self.ttc[key]
            SDL_DestroyTexture(so[1])

        self.ttc.clear()
        self.ttl.clear()

    def win_prime_font(self, name, size, user_handle, weight, y_offset=0):
    
        self.f_dict[user_handle] = Win32Font(name, int(size * self.scale), weight)
        self.y_offset_dict[user_handle] = y_offset

    def prime_font(self, name, size, user_handle, offset=0):

        self.f_dict[user_handle] = (name + " " + str(size * self.scale), offset, size * self.scale )

    def get_text_wh(self, text, font, max_x, wrap=False):

        if system == "linux":
            self.layout.set_font_description(Pango.FontDescription(self.f_dict[font][0]))
            self.layout.set_ellipsize(Pango.EllipsizeMode.END)
            self.layout.set_width(max_x * 1000)
            if wrap:
                self.layout.set_height(20000 * 1000)
            else:
                self.layout.set_height(0)
            self.layout.set_text(text, -1)

            return self.layout.get_pixel_size()
        else:
            #return self.__win_text_xy(text, font)
            return self.__win_text_xy(text, font, max_x, wrap)

    def get_y_offset(self, text, font, max_x, wrap=False):  # HACKY

        self.layout.set_font_description(Pango.FontDescription(self.f_dict[font][0]))
        self.layout.set_ellipsize(Pango.EllipsizeMode.END)
        self.layout.set_width(max_x * 1000)
        if wrap:
            self.layout.set_height(20000 * 1000)
        else:
            self.layout.set_height(0)
        self.layout.set_text(text, -1)

        y_off = self.layout.get_baseline() / 1000
        y_off = round(round(y_off) - 13 * self.scale)  # 13 for compat with way text position used to work

        return y_off

    def __render_text(self, key, x, y, range_top, range_height, align):

        sd = key

        if sd[3]:
            self.was_truncated = True

        if align == 1:
            sd[0].x = round(x) - sd[0].w

        elif align == 2:
            sd[0].x = sd[0].x - int(sd[0].w / 2)

        if range_height is not None and range_height < sd[0].h:

            if range_top < 0:
                range_top = 0

            if range_top > sd[0].h - range_height:
                range_top = sd[0].h - range_height

            self.source_rect.y = round(range_top)
            self.source_rect.w = sd[0].w
            self.source_rect.h = round(range_height)

            self.dest_rect.x = sd[0].x
            self.dest_rect.y = sd[0].y
            self.dest_rect.w = sd[0].w
            self.dest_rect.h = round(range_height)

            SDL_RenderCopyEx(self.renderer, sd[1], self.source_rect, self.dest_rect, 0, None, 0)
            return

        SDL_RenderCopy(self.renderer, sd[1], None, sd[0])


    def __draw_text_cairo(self, location, text, colour, font, max_x, bg, align=0, max_y=None, wrap=False, range_top=0, range_height=None):

        #perf.set()

        self.was_truncated = False

        max_x += 12  # Hack
        max_x = round(max_x)

        real_bg = False
        alpha_bg = self.alpha_bg

        x = round(location[0])
        y = round(location[1])

        if self.pretty_rect:

            w, h = self.get_text_wh(text, font, max_x, wrap)
            quick_box = [x, y, w, h]

            if align == 1:
                quick_box[0] = x - quick_box[2]

            elif align == 2:
                quick_box[0] = quick_box[0] - int(quick_box[2] / 2)

            if coll_rect(self.pretty_rect, quick_box):
                # self.rect_r(quick_box, [0, 0, 0, 100], True)
                # print("PT")
                # print(text)
                if self.real_bg:
                    real_bg = True
            else:
                alpha_bg = False

        if alpha_bg:
            bg = (0, 0, 0, 0)

        if max_y is not None:
            max_y = round(max_y)

        if len(text) == 0:
            return 0


        key = (max_x, text, font, colour[0], colour[1], colour[2], colour[3], bg[0], bg[1], bg[2])

        if not real_bg:
            sd = self.ttc.get(key)
            if sd:

                sd = self.ttc[key]
                sd[0].x = round(x)
                sd[0].y = round(y) - sd[2]

                self.__render_text(sd, x, y, range_top, range_height, align)

                self.ttl.remove(key)
                self.ttl.append(key)

                if wrap:
                    return sd[0].h
                return sd[0].w

        if not self.pretty_rect:  # Would have already done this if True

            w, h = self.get_text_wh(text, font, max_x, wrap)

        if w < 1:
            return 0

        h += 4  # Compensate for characters that drop past the baseline, Pango doesn't seem to allow for this

        if wrap:
            w = max_x + 1

        data = ctypes.c_buffer(b"\x00" * (h * (w * 4)))

        if real_bg:
            box = SDL_Rect(x, y - self.get_y_offset(text, font, max_x, wrap), w, h)

            if align == 1:
                box.x = x - box.w

            elif align == 2:
                box.x = box.x - int(box.w / 2)

            SDL_RenderReadPixels(self.renderer, box, SDL_PIXELFORMAT_RGB888, ctypes.pointer(data), (w * 4))

        if alpha_bg:
            surf = cairo.ImageSurface.create_for_data(data, cairo.FORMAT_ARGB32, w, h)
        else:
            surf = cairo.ImageSurface.create_for_data(data, cairo.FORMAT_RGB24, w, h)

        context = cairo.Context(surf)

        if self.force_gray:
            options = context.get_font_options()
            options.set_antialias(cairo.ANTIALIAS_GRAY)
            context.set_font_options(options)
        elif self.force_subpixel_text:
            options = context.get_font_options()
            #options.set_antialias(cairo.ANTIALIAS_NONE)
            #options.set_antialias(cairo.ANTIALIAS_GRAY)
            options.set_antialias(cairo.ANTIALIAS_SUBPIXEL)
            context.set_font_options(options)

        layout = PangoCairo.create_layout(context)
        layout.set_auto_dir(False)

        if max_y is not None:
            layout.set_ellipsize(Pango.EllipsizeMode.END)
            layout.set_width(max_x * 1000)
            layout.set_height(max_y * 1000)
        else:
            layout.set_ellipsize(Pango.EllipsizeMode.END)
            layout.set_width(max_x * 1000)

            extra = 0
            if wrap:  # Compensate for height measurement being 1-2 lines too short. Pango bug?
                extra = round(400000 * self.scale)

            layout.set_height(h * 1000 + extra)

        if not wrap and max_y is None:
            layout.set_height(-1)

        # Attributes don't seem to be implemented in gi?
        # attrs = Pango.AttrList()
        # attrs.insert(Pango.Attribute(Pango.Underline.SINGLE))
        # layout.set_attributes(attrs)

        context.rectangle(0, 0, w, h)

        if not real_bg and not alpha_bg:
            context.set_source_rgb(bg[0] / 255, bg[1] / 255, bg[2] / 255)
            # context.set_source_rgba(0, 0, 0, 0)
            context.fill()

        context.set_source_rgb(colour[0] / 255, colour[1] / 255, colour[2] / 255)

        if font not in self.f_dict:
            print("Font not loaded: " + str(font))
            return 10

        # desc = Pango.FontDescription(self.f_dict[font][0])
        # desc.set_family("Arial")

        layout.set_font_description(Pango.FontDescription(self.f_dict[font][0]))
        layout.set_text(text, -1)
        #print(layout.get_direction(0))

        y_off = layout.get_baseline() / 1000
        y_off = round(round(y_off) - 13 * self.scale)  # 13 for compat with way text position used to work

        PangoCairo.show_layout(context, layout)

        self.was_truncated = layout.is_ellipsized()

        if alpha_bg:
            sdl_surface = SDL_CreateRGBSurfaceWithFormatFrom(ctypes.pointer(data), w, h, 32, w * 4, SDL_PIXELFORMAT_ARGB8888)
        else:
            sdl_surface = SDL_CreateRGBSurfaceWithFormatFrom(ctypes.pointer(data), w, h, 24, w * 4, SDL_PIXELFORMAT_RGB888)

        # Here the background colour is keyed out allowing lines to overlap slightly
        if not real_bg and not alpha_bg:
            ke = SDL_MapRGB(sdl_surface.contents.format, bg[0], bg[1], bg[2])
            SDL_SetColorKey(sdl_surface, True, ke)

        c = SDL_CreateTextureFromSurface(self.renderer, sdl_surface)
        SDL_FreeSurface(sdl_surface)

        if alpha_bg:
            blend_mode = SDL_ComposeCustomBlendMode(SDL_BLENDFACTOR_ONE, SDL_BLENDFACTOR_ONE_MINUS_SRC_ALPHA, SDL_BLENDOPERATION_ADD, SDL_BLENDFACTOR_ONE, SDL_BLENDFACTOR_ONE_MINUS_SRC_ALPHA, SDL_BLENDOPERATION_ADD);
            SDL_SetTextureBlendMode(c, blend_mode)

        dst = SDL_Rect(round(x), round(y))
        dst.w = round(w)
        dst.h = round(h)
        dst.y = round(y) - y_off

        pack = [dst, c, y_off, self.was_truncated]

        self.__render_text(pack, x, y, range_top, range_height, align)

        # Don't cache if using real background data
        if not real_bg:
            self.ttc[key] = pack
            self.ttl.append(key)
            if len(self.ttl) > 350:
                key = self.ttl[0]
                so = self.ttc[key]
                SDL_DestroyTexture(so[1])
                del self.ttc[key]
                del self.ttl[0]
        if wrap:
            return dst.h
        return dst.w


    # WINDOWS --------------------------------------------------------

    def __win_text_xy(self, text, font, max_x, wrap):

        if font == None or font not in self.f_dict:

            print("Missing Font")
            print(font)

            return

        return self.f_dict[font].get_metrics(text, max_x, wrap)

    def __win_render_text(self, key, x, y, range_top, range_height, align):
   

        sd = key
        
        sd[0].x = round(x)
        sd[0].y = round(y)
        if align == 1:
            sd[0].x = round(x) - sd[0].w
        elif align == 2:
            sd[0].x = sd[0].x - int(sd[0].w / 2)

        if range_height is not None and range_height < sd[0].h - 20:
        
            if range_top + range_height > sd[0].h:
                # range_top = 0
                range_height = sd[0].h - range_top

            self.source_rect.y = sd[0].h - round(range_height) - round(range_top)
            self.source_rect.w = sd[0].w
            self.source_rect.h = round(range_height)

            self.dest_rect.x = sd[0].x
            self.dest_rect.y = sd[0].y
            self.dest_rect.w = sd[0].w
            self.dest_rect.h = round(range_height)
            
            SDL_RenderCopyEx(self.renderer, sd[1], self.source_rect, self.dest_rect, 0, None, SDL_FLIP_VERTICAL)
            return

        SDL_RenderCopyEx(self.renderer, sd[1], None, sd[0], 0, None, SDL_FLIP_VERTICAL)

    def __draw_text_windows(self, x, y, text, bg, fg, font=None, align=0, wrap=False, max_x=100, max_y=None, range_top=0, range_height=None):

        y += self.y_offset_dict[font]

        key = (text, font, fg[0], fg[1], fg[2], fg[3], bg[0], bg[1], bg[2], max_x)

        if key in self.cache:
            sd = self.cache[key]

            self.__win_render_text(sd, x, y, range_top, range_height, align)
            if wrap:
                return sd[0].h
            return sd[0].w

        if font is None or font not in self.f_dict:

            print("Missing Font")
            print(font)
            return 0

        #perf_timer.set()

        f = self.f_dict[font]
        
        w, h = f.get_metrics(text, max_x, wrap)
        if max_y and max_y > h:
            max_y = h

        im, c_bits = f.renderText(text, bg, fg, wrap, max_x, max_y)

        s_image = im
        ke = SDL_MapRGB(s_image.contents.format, bg[0], bg[1], bg[2])
        SDL_SetColorKey(s_image, True, ke)
        c = SDL_CreateTextureFromSurface(self.renderer, s_image)
        tex_w = pointer(c_int(0))
        tex_h = pointer(c_int(0))
        SDL_QueryTexture(c, None, None, tex_w, tex_h)
        dst = SDL_Rect(round(x), round(y))
        dst.w = int(tex_w.contents.value)
        dst.h = int(tex_h.contents.value)

        SDL_FreeSurface(s_image)
        #im.close()

        if align == 1:
            dst.x = round(x) - dst.w

        elif align == 2:
            dst.x = dst.x - int(dst.w / 2)

        #SDL_RenderCopy(renderer, c, None, dst)
        #SDL_RenderCopyEx(self.renderer, c, None, dst, 0, None, SDL_FLIP_VERTICAL)

        #print(perf_timer.get())
        self.cache[key] = [dst, c]
        self.__win_render_text([dst, c], x, y, range_top, range_height, align)

        self.ca_li.append(key)

        if len(self.ca_li) > 350:
            SDL_DestroyTexture(self.cache[self.ca_li[0]][1])
            del self.cache[self.ca_li[0]]
            del self.ca_li[0]

        return dst.w    

    def text(self, location, text, colour, font, max_w=4000, bg=None, range_top=0, range_height=None):

        #print((text, font))

        if not text:
            return 0

        max_w = max(1, max_w)

        if bg is None:
            bg = self.text_background_colour

        if colour[3] != 255:
            colour = alpha_blend(colour, bg)
        align = 0
        if len(location) > 2:
            if location[2] == 1:
                align = 1
            if location[2] == 2:
                align = 2
            if location[2] == 4:
                max_h = None
                if len(location) > 4:
                    max_h = location[4]
                    
                if system == "linux":
                    return self.__draw_text_cairo(location, text, colour, font, location[3], bg, max_y=max_h, wrap=True,
                                                  range_top=range_top, range_height=range_height)
                else:
                    return self.__draw_text_windows(location[0], location[1], text, bg, colour, font, 0, True, location[3], max_y=max_h,
                                                     range_top=range_top, range_height=range_height)

        if system == "linux":
            return self.__draw_text_cairo(location, text, colour, font, max_w, bg, align)
        else:
            return self.__draw_text_windows(location[0], location[1], text, bg, colour, font, align=align, max_x=max_w)




# Tauon Music Box - GStreamer backend Module

# Copyright © 2018-2020, Taiko2k captain(dot)gxj(at)gmail.com

#     This file is part of Tauon Music Box.
#
#     Tauon Music Box is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     Tauon Music Box is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Lesser General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with Tauon Music Box.  If not, see <http://www.gnu.org/licenses/>.


import time
import urllib.parse
import os
#import re
from t_modules.t_extra import Timer
import gi
from gi.repository import GLib

gi.require_version('Gst', '1.0')
gi.require_version('GstController', '1.0')
from gi.repository import Gst, GstController
from t_modules.t_extra import get_folder_size
import threading
import requests
import urllib.parse
import mutagen

def player3(tauon):  # GStreamer

    pctl = tauon.pctl
    lfm_scrobbler = tauon.lfm_scrobbler
    star_store = tauon.star_store
    gui = tauon.gui
    prefs = tauon.prefs

    class GPlayer:

        def __init__(self):

            print("Init GStreamer...")

            # This is used to keep track of time between callbacks.
            self.player_timer = Timer()

            # Store the track object that is currently playing
            self.loaded_track = None

            # This is used to keep note of what state of playing we should be in
            self.play_state = 0  # 0 is stopped, 1 is playing, 2 is paused

            # Initiate GSteamer
            Gst.init([])
            self.mainloop = GLib.MainLoop()

            # Populate list of output devices with defaults
            outputs = {}
            devices = ["PulseAudio", "ALSA", "JACK",]
            if tauon.snap_mode:  # Snap permissions don't support these by default
                devices.remove("JACK")
                devices.remove("ALSA")

            # Get list of available audio device
            self.dm = Gst.DeviceMonitor()
            self.dm.start()
            for device in self.dm.get_devices():
                if device.get_device_class() == "Audio/Sink":
                    element = device.create_element(None)
                    type_name = element.get_factory().get_name()
                    if hasattr(element.props, "device"):
                        device_name = element.props.device
                        display_name = device.get_display_name()

                        # This is used by the UI to present list of options to the user in audio settings
                        outputs[display_name] = (type_name, device_name)
                        devices.append(display_name)

            # dm.stop()  # Causes a segfault sometimes
            pctl.gst_outputs = outputs
            pctl.gst_devices = devices

            # Create main "playbin" pipeline for playback
            self.playbin = Gst.ElementFactory.make("playbin", "player")

            # Create custom output bin from user preferences
            if not prefs.gst_use_custom_output:
                prefs.gst_output = prefs.gen_gst_out()

            self._output = Gst.parse_bin_from_description(
                prefs.gst_output, ghost_unlinked_pads=True)

            # Create a bin for the audio pipeline
            self._sink = Gst.ElementFactory.make("bin", "sink")
            self._sink.add(self._output)

            # Spectrum -------------------------
            # Cant seem to figure out how to process these magnitudes in a way that looks good

            # self.spectrum = Gst.ElementFactory.make("spectrum", "spectrum")
            # self.spectrum.set_property('bands', 20)
            # self.spectrum.set_property('interval', 10000000)
            # self.spectrum.set_property('threshold', -100)
            # self.spectrum.set_property('post-messages', True)
            # self.spectrum.set_property('message-magnitude', True)
            #
            # self.playbin.set_property('audio-filter', self.spectrum)
            # # ------------------------------------

            # # Level Meter -------------------------
            self.level = Gst.ElementFactory.make("level", "level")
            self.level.set_property('interval', 20000000)
            self.playbin.set_property('audio-filter', self.level)
            # # ------------------------------------

            self._eq = Gst.ElementFactory.make("equalizer-nbands", "eq")
            self._vol = Gst.ElementFactory.make("volume", "volume")

            self._sink.add(self._eq)
            self._sink.add(self._vol)

            self._eq.link(self._vol)
            self._vol.link(self._output)

            self._eq.set_property("num-bands", 10 + 2)

            # Set the equalizer based on user preferences

            # Using workaround for "inverted slider" bug.
            # Thanks to Lollypop and Clementine for discovering this.
            # Ref https://github.com/Taiko2k/TauonMusicBox/issues/414

            for i in range(10 + 2):
                band = self._eq.get_child_by_index(i)
                band.set_property("freq", 0)
                band.set_property("bandwidth", 0)
                band.set_property("gain", 0)

            self.set_eq()
            # if prefs.use_eq:
            #     self._eq.set_property("band" + str(i), level * -1)
            # else:
            #     self._eq.set_property("band" + str(i), 0.0)

            # Set up sink pad for the intermediate bin via the
            # first element (volume)
            ghost = Gst.GhostPad.new("sink", self._eq.get_static_pad("sink"))
            self._sink.add_pad(ghost)

            # Connect the playback bin to to the intermediate bin sink pad
            self.playbin.set_property("audio-sink", self._sink)

            # The pipeline should look something like this -
            # (player) -> [(eq) -> (volume) -> (output)]

            # Create controller for pause/resume volume fading
            self.c_source = GstController.InterpolationControlSource()
            self.c_source.set_property('mode', GstController.InterpolationMode.LINEAR)
            self.c_binding = GstController.DirectControlBinding.new(self._vol, "volume", self.c_source)
            self._vol.add_control_binding(self.c_binding)

            # Set callback for the main callback loop
            GLib.timeout_add(50, self.main_callback)

            self.playbin.connect("about-to-finish", self.about_to_finish)  # Not used

            # Setup bus and select what types of messages we want to listen for
            bus = self.playbin.get_bus()
            bus.add_signal_watch()
            bus.connect('message::element', self.on_message)
            bus.connect('message::buffering', self.on_message)
            bus.connect('message::error', self.on_message)
            bus.connect('message::tag', self.on_message)
            bus.connect('message::warning', self.on_message)
            # bus.connect('message::eos', self.on_message)

            # Variables used with network downloading
            self.temp_id = "a"
            self.url = None
            self.dl_ready = True
            self.using_cache = False
            self.temp_path = ""  # Full path + filename
            # self.level_train = []
            self.seek_timer = Timer()
            self.seek_timer.force_set(10)
            self.buffering = False
            # Other
            self.end_timer = Timer()

            # Start GLib mainloop
            self.mainloop.run()

        def set_eq(self):
            last = 0
            for i, level in enumerate(prefs.eq):
                band = self._eq.get_child_by_index(i + 1)
                if prefs.use_eq:
                    band = self._eq.get_child_by_index(i + 1)
                    freq = 31.25 * (2 ** i)
                    band.set_property("freq", freq)
                    band.set_property("bandwidth", freq - last)
                    last = freq
                    band.set_property("gain", level * -1)
                else:
                    band.set_property("freq", 0)
                    band.set_property("bandwidth", 0)
                    band.set_property("gain", 0)

        def about_to_finish(self, bin):
            self.end_timer.set()

        def on_message(self, bus, msg):
            struct = msg.get_structure()
            #print(struct.get_name())
            #print(struct.to_string())

            name = struct.get_name()

            if name == "GstMessageError":

                if "_is_dead" in struct.to_string():
                    # Looks like PulseAudio was reset. Need to restart playback.

                    self.playbin.set_state(Gst.State.NULL)

                    if tauon.stream_proxy.download_running:
                        tauon.stream_proxy.stop()
                    else:
                        self.playbin.set_state(Gst.State.PLAYING)
                        tries = 0
                        while tries < 10:
                            time.sleep(0.03)
                            r = self.playbin.seek_simple(Gst.Format.TIME, Gst.SeekFlags.FLUSH | Gst.SeekFlags.KEY_UNIT,
                                                         (pctl.start_time_target + pctl.playing_time) * Gst.SECOND)
                            if r:
                                break
                            tries += 1

            if self.play_state == 3 and name == "GstMessageTag":

                if not tauon.radiobox.parse_vorbis_okay():
                    return

                data = struct.get_value("taglist").get_string("title")
                data2 = struct.get_value("taglist").get_string("artist")
                data3 = struct.get_value("taglist").get_string("year")
                data4 = struct.get_value("taglist").get_string("album")
                # print(struct.to_string())
                if data[0]:
                    pctl.tag_meta = ""
                    line = ""
                    line = data[1]
                    if data2[0]:
                        line = data2[1] + " - " + line

                    pctl.found_tags = {}

                    pctl.found_tags["title"] = data[1]
                    if data2[0]:
                        pctl.found_tags["artist"] = data2[1]
                    if data3[0]:
                        pctl.found_tags["year"] = str(data3[1])
                    if data4[0]:
                        pctl.found_tags["album"] = data4[1]

                    pctl.tag_meta = line
                    print("Found tag: " + line)

            elif name == "GstMessageError":
                if "Connection" in struct.get_value("debug"):
                    gui.show_message("Connection error", mode="info")
            elif name == 'GstMessageBuffering':

                if pctl.playing_state == 3:
                    buff_percent = struct.get_value("buffer-percent")

                    if buff_percent == 0 and (self.play_state == 1 or self.play_state == 3):
                        self.playbin.set_state(Gst.State.PAUSED)
                        self.buffering = True
                        print("Buffering...")

                    elif self.buffering and buff_percent == 100 and (self.play_state == 1 or self.play_state == 3):
                        self.playbin.set_state(Gst.State.PLAYING)
                        self.buffering = False
                        print("Buffered")

            if gui.vis == 1 and name == 'level':
                # print(struct.to_string())
                data = struct.get_value("peak")
                ts = struct.get_value("timestamp")
                # print(data)
                l = (10 ** (data[0] / 20)) * 11.6
                if len(data) == 1:
                    r = l
                else:
                    r = (10 ** (data[1] / 20)) * 11.6

                td = (ts / 1000000000) - (self.playbin.query_position(Gst.Format.TIME)[1] / Gst.SECOND)
                t = time.time()
                rt = t + td
                if td > 0:
                    for item in tauon.level_train:
                        if rt < item[0]:
                            tauon.level_train.clear()
                            # print("FF")
                            break
                    tauon.level_train.append((rt, l, r))

            # if name == 'spectrum':
            #     struct_str = struct.to_string()
            #     magnitude_str = re.match(r'.*magnitude=\(float\){(.*)}.*', struct_str)
            #     if magnitude_str:
            #         magnitude = map(float, magnitude_str.group(1).split(','))
            #
            #         l = list(magnitude)
            #         k = []
            #         #print(l)
            #         for a in l[0:20]:
            #             #v = ??
            #             k.append()
            #         print(k)
            #         gui.spec = k
            #         #print(k)
            #         gui.level_update = True
            #
            # return True


        def check_duration(self):

            # This function is to be called when loading a track to query for a duration of track
            # in case the tagger failed to calculate a length for the track when imported.

            # Get current playing track object from player
            current_track = pctl.playing_object()

            if current_track is not None and (current_track.length < 1 or current_track.file_ext.lower() in tauon.mod_formats):

                time.sleep(0.25)

                result = self.playbin.query_duration(Gst.Format.TIME)

                if result[0] is True:
                    current_track.length = result[1] / Gst.SECOND
                    pctl.playing_length = current_track.length
                    gui.pl_update += 1

                else:  # still loading? I guess we wait and try again.
                    time.sleep(1.5)
                    result = self.playbin.query_duration(Gst.Format.TIME)

                    if result[0] is True:
                        current_track.length = result[1] / Gst.SECOND

        def download_part(self, url, target, params, id):

            #   GStreamer can't seek some types of HTTP sources.
            #
            #   To work around this, when a seek is requested by the user, this
            #   function facilitates the download of the URL in whole, then loaded
            #   into GStreamer as a complete file to provide at least some manner of seeking
            #   ability for the user. (User must wait for full download)
            #
            #   (Koel and Airsonic MP3 sources are exempt from this as seeking does work with them)
            #
            #   A better solution might be to download file externally then feed the audio data
            #   into GStreamer as it downloads. This still would have the issue that the whole file
            #   must have been downloaded before a seek could begin.
            #
            #   With the old BASS backend, this was done with the file on disk being constantly
            #   appended to. Unfortunately GStreamer doesn't support playing files in this manner.

            try:
                part = requests.get(url, stream=True, params=params)
            except:
                gui.show_message("Could not connect to server", mode="error")
                self.dl_ready = True
                return

            bitrate = 0

            a = 0
            z = 0
            # print(target)
            f = open(target, "wb")
            for chunk in part.iter_content(chunk_size=1024):
                if chunk:  # filter out keep-alive new chunks
                    a += 1
                    # if a == 300:  # kilobyes~
                    #     self.dl_ready = True

                    if id != self.id:
                        part.close()
                        f.close()
                        os.remove(target)
                        return
                        # break

                    f.write(chunk)

                    # Periodically update download the progress indicator
                    z += 1
                    if id == self.id:
                        if z == 60:
                            z = 0
                            if bitrate == 0:
                                try:
                                    audio = mutagen.File(target)
                                    bitrate = audio.info.bitrate / 1000
                                except:
                                    pass

                            if bitrate > 0:
                                gui.update += 1
                                pctl.download_time = a * 1024 / (bitrate / 8) / 1000
            f.close()
            #pctl.download_time = -1

            self.dl_ready = True

        def main_callback(self):

            if not pctl.playerCommandReady and pctl.playing_state == 0 and not tauon.spot_ctl.playing and not tauon.spot_ctl.coasting:
                # tauon.tm.player_lock.acquire()  # Blocking the GLib thread blocks tray processing
                GLib.timeout_add(50, self.main_callback)
                return

            if gui.vis == 1:
                if pctl.playing_state == 1:
                    gui.level_update = True

            # This is the main callback function to be triggered continuously as long as application is running
            if self.play_state == 1 and pctl.playing_time > 1 and not pctl.playerCommandReady:

                pctl.test_progress()  # This function triggers an advance if we are near end of track

                success, state, pending = self.playbin.get_state(3 * Gst.SECOND)
                if state != Gst.State.PLAYING:
                    time.sleep(0.5)

                    print("Stall...")

            if self.play_state == 3:
                pctl.radio_progress()

            if not pctl.playerCommandReady:
                pctl.spot_test_progress()

            if pctl.playerCommandReady:
                command = pctl.playerCommand
                pctl.playerCommand = ""
                pctl.playerCommandReady = False


                # print("RUN COMMAND")
                # print(command)

                # Here we process commands from the main thread/module

                # Possible commands:

                # open:     Start playback of a file
                #           Path given by pctl.target_open at position pctl.start_time_target + pctl.jump_time
                # stop:     Stop playback (Implies release file)
                # runstop:  Stop playback but let finish if we are near the end of the file
                # pauseon:  Pause playback (be ready to resume)
                # pauseoff: Resume playback if paused
                # volume:   Set to the volume specified by pctl.player_volume (0 to 100)
                # seek:     Seek to position given by pctl.new_time + pctl.start_time (don't resume playback if paused)
                # url:      Start playback of a shoutcast/icecast stream. URL specified by pctl.url
                #           todo: start recording if pctl.record_stream  (rec button is current disabled for GST in UI)
                #                 encode to OGG and output file to prefs.encoder_output
                #                 automatically name files and split on metadata change
                # unload:   Stop, cleanup and exit thread
                # done:     Tell the main thread we finished doing a special request it was waiting for (such as unload)

                # Todo: Visualisers (Hard)
                # Ideally we would want the same visual effect as the old BASS based visualisers.
                # What we want to do is constantly get binned spectrum data and pass it to the UI.
                # Specifically, format used with BASS module is:
                # - An FFT of sample data with Hanning window applied
                # - 1024 samples (returns first half, 512 values)
                # - Non-complex (magnitudes)
                # - Combined left and right channels
                # - Binned to particular numbers of bars and passed onto UI after some scaling and truncating

                #pctl.download_time = 0
                url = None
                if command == 'open' and pctl.target_object:
                    # print("Start track")
                    track = pctl.target_object

                    if (tauon.spot_ctl.playing or tauon.spot_ctl.coasting) and not track.file_ext == "SPTY":
                        tauon.spot_ctl.control("stop")

                    if tauon.stream_proxy.download_running:
                        tauon.stream_proxy.stop()

                    # Check if the file exists, mark it as missing if not
                    if track.is_network:

                        if track.file_ext == "SPTY":
                            tauon.level_train.clear()
                            if self.play_state > 0:
                                self.playbin.set_state(Gst.State.READY)
                            self.play_state = 0
                            try:
                                tauon.spot_ctl.play_target(track.url_key)
                            except:
                                print("Failed to start Spotify track")
                                pctl.playerCommand = "stop"
                                pctl.playerCommandReady = True

                            GLib.timeout_add(19, self.main_callback)
                            return

                        try:
                            url, params = pctl.get_url(track)
                            self.urlparams = url, params
                        except:
                            time.sleep(0.1)
                            gui.show_message("Connection error", "Bad login? Server offline?", mode='info')
                            pctl.stop()
                            pctl.playerCommand = ""
                            self.main_callback()
                            return

                    # If the target is a file, check that is exists
                    elif os.path.isfile(track.fullpath):
                        if not track.found:
                            pctl.reset_missing_flags()
                    else:
                        # File does not exist, force trigger an advance
                        track.found = False
                        tauon.console.print("Missing File: " + track.fullpath, 2)
                        pctl.playing_state = 0
                        pctl.jump_time = 0
                        # print("FORCE JUMP")
                        pctl.advance(inplace=True, play=True)
                        GLib.timeout_add(19, self.main_callback)
                        return

                    gapless = False
                    current_time = 0
                    current_duration = 0

                    if track.is_network:
                        if params:
                            self.url = url + ".view?" + urllib.parse.urlencode(params)
                        else:
                            self.url = url

                    if self.play_state != 0:
                        # Determine time position of currently playing track
                        current_time = self.playbin.query_position(Gst.Format.TIME)[1] / Gst.SECOND
                        current_duration = self.playbin.query_duration(Gst.Format.TIME)[1] / Gst.SECOND
                        # print("We are " + str(current_duration - current_time) + " seconds from end.")

                    if self.play_state != 1:
                        self.loaded_track = None

                    # If we are close to the end of the track, try transition gaplessly
                    if self.play_state == 1 and pctl.start_time_target == 0 and pctl.jump_time == 0 and \
                            current_duration - current_time < 5.5 and not pctl.playerSubCommand == 'now' \
                            and self.end_timer.get() > 3:

                        gapless = True

                        # if self.play_state == 1 and self.loaded_track and self.loaded_track.is_network:
                        #     # Gst may report wrong length for network tracks, use known length instead
                        #     if pctl.playing_time < self.loaded_track.length - 4:
                        #         gapless = False


                    # We're not at the end of the last track so reset the pipeline
                    if not gapless:
                        self.playbin.set_state(Gst.State.NULL)
                        tauon.level_train.clear()

                    pctl.playerSubCommand = ""
                    self.play_state = 1

                    self.save_temp = tauon.temp_audio + "/" + str(track.index) + "-audio"
                    # shoot_dl = threading.Thread(target=self.download_part,
                    #                             args=([url, self.save_temp, params, track.url_key]))
                    # shoot_dl.daemon = True
                    # shoot_dl.start()
                    self.using_cache = False

                    self.id = track.url_key
                    if url:
                        # self.playbin.set_property('uri',
                        #                           'file://' + urllib.parse.quote(os.path.abspath(self.save_temp)))

                        if self.dl_ready and os.path.exists(self.save_temp):
                            self.using_cache = True
                            self.playbin.set_property('uri',
                                                      'file://' + urllib.parse.quote(os.path.abspath(self.save_temp)))
                        else:
                            self.playbin.set_property('uri', self.url)
                    else:
                        # Play file on disk
                        self.playbin.set_property('uri', 'file://' + urllib.parse.quote(os.path.abspath(track.fullpath)))

                    if pctl.start_time_target > 0:
                        self.playbin.set_property('volume', 0.0)
                    else:
                        self.playbin.set_property('volume', pctl.player_volume / 100)

                    self.playbin.set_state(Gst.State.PLAYING)
                    if pctl.jump_time == 0 and not pctl.playerCommand == "seek":
                        pctl.playing_time = 0

                    # The position to start is not always the beginning of the file, so seek to position
                    if pctl.start_time_target > 0 or pctl.jump_time > 0:

                        tries = 0
                        while tries < 150:
                            time.sleep(0.03)
                            r = self.playbin.seek_simple(Gst.Format.TIME, Gst.SeekFlags.FLUSH | Gst.SeekFlags.KEY_UNIT,
                                                (pctl.start_time_target + pctl.jump_time) * Gst.SECOND)
                            if r:
                                break
                            tries += 1
                            if tries > 2:
                                print("Seek failed, retrying...")
                                print(tries)

                        pctl.playing_time = 0
                        gui.update = 1

                        self.playbin.set_property('volume', pctl.player_volume / 100)

                    if gapless:  # Hold thread while a gapless transition is in progress
                        t = 0
                        # print("Gapless go")
                        while self.playbin.query_position(Gst.Format.TIME)[1] / Gst.SECOND >= current_time > 0:

                            time.sleep(0.02)
                            t += 1

                            if self.playbin.get_state(0).state != Gst.State.PLAYING:
                                break

                            if t > 250:
                                print("Gonna stop waiting...")  # Cant wait forever
                                break

                            if pctl.playerCommand == 'open' and pctl.playerCommandReady:
                                # Cancel the gapless transition
                                print("Cancel transition")
                                self.playbin.set_state(Gst.State.NULL)
                                time.sleep(0.5)
                                GLib.timeout_add(19, self.main_callback)
                                return


                    add_time = max(min(self.player_timer.hit(), 3), 0)
                    if self.loaded_track and self.loaded_track in pctl.master_library.values():
                        star_store.add(self.loaded_track.index, add_time)

                    self.loaded_track = track

                    pctl.jump_time = 0
                    #time.sleep(1)
                    add_time = self.player_timer.hit()
                    if add_time > 2:
                        add_time = 2
                    if add_time < 0:
                        add_time = 0
                    pctl.playing_time += add_time

                    t = self.playbin.query_position(Gst.Format.TIME)
                    if t[0]:
                        pctl.decode_time = (t[1] / Gst.SECOND) - self.loaded_track.start_time
                    else:
                        pctl.decode_time = pctl.playing_time

                    if self.loaded_track:
                        star_store.add(self.loaded_track.index, add_time)


                    self.check_duration()
                    self.player_timer.hit()
                    time.sleep(0.5)

                elif command == 'url':

                    # Stop if playing or paused
                    if self.play_state == 1 or self.play_state == 2 or self.play_state == 3:
                        self.playbin.set_state(Gst.State.NULL)
                        time.sleep(0.1)

                    w = 0
                    while len(tauon.stream_proxy.chunks) < 50:
                        time.sleep(0.01)
                        w += 1
                        if w > 500:
                            print("Taking too long!")
                            tauon.stream_proxy.stop()
                            pctl.playerCommand = 'stop'
                            pctl.playerCommandReady = True
                            break
                    else:
                        # Open URL stream
                        self.playbin.set_property('uri', pctl.url)
                        self.playbin.set_property('volume', pctl.player_volume / 100)
                        self.buffering = False
                        self.playbin.set_state(Gst.State.PLAYING)
                        self.play_state = 3
                        self.player_timer.hit()

                elif command == 'seteq':
                    self.set_eq()

                    # for i, level in enumerate(prefs.eq):
                    #     if prefs.use_eq:
                    #         self._eq.set_property("band" + str(i), level * -1)
                    #     else:
                    #         self._eq.set_property("band" + str(i), 0.0)

                elif command == 'volume':

                    if tauon.spot_ctl.coasting or tauon.spot_ctl.playing:
                        tauon.spot_ctl.control("volume", int(pctl.player_volume))

                    elif self.play_state == 1 or self.play_state == 3:

                        success, current_time = self.playbin.query_position(Gst.Format.TIME)
                        self.playbin.set_state(Gst.State.PLAYING)

                        if success and False:
                            start = current_time + ((150 / 1000) * Gst.SECOND)
                            end = current_time + ((200 / 1000) * Gst.SECOND)
                            self.c_source.set(start, self._vol.get_property('volume') / 10)
                            self.c_source.set(end, (pctl.player_volume / 100) / 10)
                            time.sleep(0.6)
                            self.c_source.unset_all()
                        else:
                            self.playbin.set_property('volume', pctl.player_volume / 100)

                elif command == 'runstop':

                    if self.play_state != 0:
                        # Determine time position of currently playing track
                        current_time = self.playbin.query_position(Gst.Format.TIME)[1] / Gst.SECOND
                        current_duration = self.playbin.query_duration(Gst.Format.TIME)[1] / Gst.SECOND
                        if current_duration - current_time < 5.5:
                            time.sleep((current_duration - current_time) + 1)
                            self.playbin.set_state(Gst.State.READY)
                        else:
                            self.playbin.set_state(Gst.State.READY)
                    else:
                        self.playbin.set_state(Gst.State.READY)
                    self.play_state = 0
                    pctl.playerSubCommand = "stopped"

                elif command == 'stop':
                    if self.play_state > 0:

                        if prefs.use_pause_fade:
                            success, current_time = self.playbin.query_position(Gst.Format.TIME)
                            if success:
                                start = current_time + (150 / 1000 * Gst.SECOND)
                                end = current_time + (prefs.pause_fade_time / 1000 * Gst.SECOND)
                                self.c_source.set(start, (pctl.player_volume / 100) / 10)
                                self.c_source.set(end, 0.0)
                                time.sleep(prefs.pause_fade_time / 1000)
                                time.sleep(0.05)
                                self.c_source.unset_all()

                        self.playbin.set_state(Gst.State.NULL)
                        time.sleep(0.1)
                        self._vol.set_property("volume", pctl.player_volume / 100)

                    self.play_state = 0
                    pctl.playerSubCommand = "stopped"

                elif command == 'seek':
                    self.seek_timer.set()
                    if tauon.spot_ctl.coasting or tauon.spot_ctl.playing:
                        tauon.spot_ctl.control("seek", int(pctl.new_time * 1000))
                        pctl.playing_time = pctl.new_time

                    elif self.play_state > 0:
                        if not self.using_cache and pctl.target_object.is_network and \
                                not pctl.target_object.file_ext == "KOEL" and \
                                not pctl.target_object.file_ext == "SUB" and \
                                not pctl.target_object.file_ext == "JELY":

                            if not os.path.exists(tauon.temp_audio):
                                os.makedirs(tauon.temp_audio)

                            listing = os.listdir(tauon.temp_audio)
                            full = [os.path.join(tauon.temp_audio, x) for x in listing]
                            size = get_folder_size(tauon.temp_audio) / 1000000
                            print(f"Audio cache size is {size}MB")
                            if size > 120:
                                oldest_file = min(full, key=os.path.getctime)
                                print("Cache full, delete oldest cached file")
                                os.remove(oldest_file)

                            pctl.playing_time = 0
                            self.playbin.set_state(Gst.State.NULL)
                            self.dl_ready = False
                            url, params = self.urlparams
                            shoot_dl = threading.Thread(target=self.download_part,
                                                        args=([url, self.save_temp, params, pctl.target_object.url_key]))
                            shoot_dl.daemon = True
                            shoot_dl.start()
                            pctl.playerCommand = ""
                            while not self.dl_ready:
                                # print("waiting...")
                                time.sleep(0.25)
                                if pctl.playerCommandReady and pctl.playerCommand != "seek":
                                    print("BREAK!")
                                    self.main_callback()
                                    return

                            self.playbin.set_property('uri', 'file://' + urllib.parse.quote(
                                os.path.abspath(self.save_temp)))
                            #time.sleep(0.05)
                            self.playbin.set_state(Gst.State.PLAYING)
                            self.using_cache = True
                            time.sleep(0.1)

                        self.playbin.seek_simple(Gst.Format.TIME, Gst.SeekFlags.FLUSH | Gst.SeekFlags.KEY_UNIT,
                                                 (pctl.new_time + pctl.start_time_target) * Gst.SECOND)


                        # It may take a moment for seeking to update when streaming, so for better UI feedback we'll
                        # update the seek indicator immediately and hold the thread for a moment
                        if pctl.target_object.is_network:
                            pctl.playing_time = pctl.new_time + pctl.start_time_target
                            pctl.decode_time = pctl.playing_time
                            time.sleep(0.25)


                elif command == 'pauseon':
                    self.player_timer.hit()
                    self.play_state = 2

                    if prefs.use_pause_fade:
                        success, current_time = self.playbin.query_position(Gst.Format.TIME)
                        if success:
                            start = current_time + (150 / 1000 * Gst.SECOND)
                            end = current_time + (prefs.pause_fade_time / 1000 * Gst.SECOND)
                            self.c_source.set(start, (pctl.player_volume / 100) / 10)
                            self.c_source.set(end, 0.0)
                            time.sleep(prefs.pause_fade_time / 1000)
                            self.c_source.unset_all()

                    self.playbin.set_state(Gst.State.PAUSED)

                elif command == 'pauseoff':
                    self.player_timer.hit()

                    if not prefs.use_pause_fade:
                        self.playbin.set_state(Gst.State.PLAYING)

                    else:
                        self._vol.set_property("volume", 0.0)
                        success, current_time = self.playbin.query_position(Gst.Format.TIME)
                        self.playbin.set_state(Gst.State.PLAYING)
                        if success:
                            start = current_time + (150 / 1000 * Gst.SECOND)
                            end = current_time + ((prefs.pause_fade_time / 1000) * Gst.SECOND)
                            self.c_source.set(start, 0.0)
                            self.c_source.set(end, (pctl.player_volume / 100) / 10)
                            time.sleep(prefs.pause_fade_time / 1000)
                            time.sleep(0.05)
                            self.c_source.unset_all()

                    self._vol.set_property("volume", pctl.player_volume / 100)


                    self.play_state = 1

                elif command == 'unload':
                    if self.play_state > 0:
                        self.playbin.set_state(Gst.State.NULL)
                        time.sleep(0.05)
                    print("Unload GStreamer")
                    self.mainloop.quit()
                    pctl.playerCommand = 'done'
                    return

            if self.play_state == 3:
                if self.playbin.get_state(0).state == Gst.State.PLAYING:
                    add_time = self.player_timer.hit()
                    if add_time > 2:
                        add_time = 2
                    if add_time < 0:
                        add_time = 0
                    pctl.playing_time += add_time
                    pctl.decode_time = pctl.playing_time

            if self.play_state == 1:

                # Get jump in time since last call
                add_time = self.player_timer.hit()

                # Limit the jump.
                if add_time > 2:
                    add_time = 2
                if add_time < 0:
                    add_time = 0

                # # Progress main seek head
                if self.playbin.get_state(0).state == Gst.State.PLAYING and self.seek_timer.get() > 1 and not pctl.playerCommandReady:

                    pctl.playing_time += add_time

                    p = max(0, (self.playbin.query_position(Gst.Format.TIME)[1] / Gst.SECOND) -
                                                 pctl.start_time_target)

                    if abs(pctl.playing_time - p) > 0.2:
                        pctl.playing_time = p

                    pctl.decode_time = p

                else:
                    # We're supposed to be playing but it's not? Give it a push I guess.
                    #self.playbin.set_state(Gst.State.PLAYING)
                    pctl.playing_time += add_time
                    pctl.decode_time = pctl.playing_time

                # Other things we need to progress such as scrobbling
                if pctl.playing_time < 3 and pctl.a_time < 3:
                    pctl.a_time = pctl.playing_time
                else:
                    pctl.a_time += add_time

                pctl.total_playtime += add_time
                lfm_scrobbler.update(add_time)  # This handles other scrobblers such as listenbrainz also

                # Update track total playtime
                if len(pctl.track_queue) > 0 and 2 > add_time > 0:
                    star_store.add(pctl.track_queue[pctl.queue_step], add_time)

            if not pctl.running:
                # print("unloading gstreamer")
                if self.play_state > 0:
                    self.playbin.set_state(Gst.State.NULL)
                    time.sleep(0.5)

                self.mainloop.quit()
                pctl.playerCommand = 'done'

            else:
                if gui.vis == 1:
                    GLib.timeout_add(19, self.main_callback)
                else:
                    GLib.timeout_add(50, self.main_callback)

        def exit(self):
            print("GStreamer unloaded")
            pctl.playerCommand = 'done'

    player = GPlayer()

    # try:
    #     player.dm.stop()
    # except:
    #     pass

    # Notify main thread we have closed cleanly
    player.exit()



# Tauon Music Box - Theme reading module

# Copyright © 2015-2020, Taiko2k captain(dot)gxj(at)gmail.com

#     This file is part of Tauon Music Box.
#
#     Tauon Music Box is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     Tauon Music Box is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Lesser General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with Tauon Music Box.  If not, see <http://www.gnu.org/licenses/>.


def get_colour_from_line(cline):

    colour = ["", "", "", ""]
    mode = 0

    for i in cline:
        if i.isdigit():
            colour[mode] += i
        elif i == ',':
            mode += 1

    for b in range(len(colour)):
        if colour[b] == "":
            colour[b] = "255"
        colour[b] = int(colour[b])

    return colour


def load_theme(colours, path):

    f = open(path, encoding="utf-8")
    content = f.readlines()

    for p in content:
        if "#" in p:
            continue
        if "light-mode" in p:
            colours.light_mode()
        if 'window frame' in p:
            colours.window_frame = get_colour_from_line(p)
        if 'gallery highlight' in p:
            colours.gallery_highlight = get_colour_from_line(p)
        if 'index playing' in p:
            colours.index_playing = get_colour_from_line(p)
        if 'time playing' in p:
            colours.time_text = get_colour_from_line(p)
        if 'artist playing' in p:
            colours.artist_playing = get_colour_from_line(p)
        if 'album line' in p:  # Bad name
            colours.album_text = get_colour_from_line(p)
        if 'track album' in p:
            colours.album_text = get_colour_from_line(p)
        if 'album playing' in p:
            colours.album_playing = get_colour_from_line(p)
        if 'player background' in p:  # bad name
            colours.top_panel_background = get_colour_from_line(p)
        if 'top panel' in p:
            colours.top_panel_background = get_colour_from_line(p)
        if 'queue panel' in p:
            colours.queue_background = get_colour_from_line(p)
        if 'side panel' in p:
            colours.side_panel_background = get_colour_from_line(p)
            colours.playlist_box_background = colours.side_panel_background
        if 'gallery background' in p:
            colours.gallery_background = get_colour_from_line(p)
        if 'playlist panel' in p:  # bad name
            colours.playlist_panel_background = get_colour_from_line(p)
        if 'tracklist panel' in p:
            colours.playlist_panel_background = get_colour_from_line(p)
        if 'track line' in p:
            colours.title_text = get_colour_from_line(p)
        if 'track missing' in p:
            colours.playlist_text_missing = get_colour_from_line(p)
        if 'playing highlight' in p:
            colours.row_playing_highlight = get_colour_from_line(p)
        if 'track time' in p:
            colours.bar_time = get_colour_from_line(p)
        if 'fav line' in p:
            colours.star_line = get_colour_from_line(p)
        if 'folder title' in p:
            colours.folder_title = get_colour_from_line(p)
        if 'folder line' in p:
            colours.folder_line = get_colour_from_line(p)
        if 'buttons off' in p:
            colours.media_buttons_off = get_colour_from_line(p)
        if 'buttons over' in p:
            colours.media_buttons_over = get_colour_from_line(p)
        if 'buttons active' in p:
            colours.media_buttons_active = get_colour_from_line(p)
        if 'playing time' in p:
            colours.time_playing = get_colour_from_line(p)
        if 'track index' in p:
            colours.index_text = get_colour_from_line(p)
        if 'track playing' in p:
            colours.title_playing = get_colour_from_line(p)
        if 'select highlight' in p:
            colours.row_select_highlight = get_colour_from_line(p)
        if 'track artist' in p:
            colours.artist_text = get_colour_from_line(p)
        if 'tab active line' in p:  # bad name
            colours.tab_text_active = get_colour_from_line(p)
        if 'tab line' in p:  # bad name
            colours.tab_text = get_colour_from_line(p)
        if 'tab active text' in p:
            colours.tab_text_active = get_colour_from_line(p)
        if 'tab text' in p:
            colours.tab_text = get_colour_from_line(p)
        if 'tab background' in p:
            colours.tab_background = get_colour_from_line(p)
        if 'tab over' in p:
            colours.tab_highlight = get_colour_from_line(p)
        if 'tab active background' in p:
            colours.tab_background_active = get_colour_from_line(p)
        if 'title info' in p:
            colours.side_bar_line1 = get_colour_from_line(p)
        if 'extra info' in p:
            colours.side_bar_line2 = get_colour_from_line(p)
        if 'bottom title' in p:
            colours.bar_title_text = get_colour_from_line(p)
        if 'scroll bar' in p:
            colours.scroll_colour = get_colour_from_line(p)
        if 'seek bar' in p:
            colours.seek_bar_fill = get_colour_from_line(p)
        if 'seek bg' in p:
            colours.seek_bar_background = get_colour_from_line(p)
        if 'volume bar' in p:
            colours.volume_bar_fill = get_colour_from_line(p)
        if 'volume bg' in p:
            colours.volume_bar_background = get_colour_from_line(p)
        if 'mode off' in p:
            colours.mode_button_off = get_colour_from_line(p)
        if 'mode over' in p:
            colours.mode_button_over = get_colour_from_line(p)
        if 'mode on' in p:
            colours.mode_button_active = get_colour_from_line(p)
        if 'art border' in p:
            colours.art_box = get_colour_from_line(p)
        if 'tb line' in p:
            colours.tb_line = get_colour_from_line(p)
        if 'music vis' in p:
            colours.vis_colour = get_colour_from_line(p)
        if 'menu background' in p:
            colours.menu_background = get_colour_from_line(p)
        if 'menu text' in p:
            colours.menu_text = get_colour_from_line(p)
        if 'menu disable' in p:
            colours.menu_text_disabled = get_colour_from_line(p)
        if 'menu icons' in p:
            colours.menu_icons = get_colour_from_line(p)
        if 'menu highlight' in p:
            colours.menu_highlight_background = get_colour_from_line(p)
        if 'menu border' in p:
            colours.menu_tab = get_colour_from_line(p)
        if 'lyrics showcase' in p:
            colours.lyrics = get_colour_from_line(p)
        if 'bottom panel' in p:
            colours.bottom_panel_colour = get_colour_from_line(p)
            # colours.menu_background = colours.bottom_panel_colour
        if 'mini bg' in p:
            colours.mini_mode_background = get_colour_from_line(p)
        if 'mini border' in p:
            colours.mini_mode_border = get_colour_from_line(p)
        if 'mini text 1' in p:
            colours.mini_mode_text_1 = get_colour_from_line(p)
        if 'mini text 2' in p:
            colours.mini_mode_text_2 = get_colour_from_line(p)
        if 'column-' in p:
            key = p[p.find("column-") + 7:].replace("-", " ").lower().title().rstrip()
            value = get_colour_from_line(p)
            colours.column_colours[key] = value
        if 'column+' in p:
            key = p[p.find("column+") + 7:].replace("-", " ").lower().title().rstrip()
            value = get_colour_from_line(p)
            colours.column_colours_playing[key] = value
        if 'menu bg' in p:
            colours.menu_background = get_colour_from_line(p)
        if 'playlist box bg' in p:  # bad name
            colours.playlist_box_background = get_colour_from_line(p)
        if 'playlist background' in p:
            colours.playlist_box_background = get_colour_from_line(p)

        if 'box background' in p:
            colours.box_background = get_colour_from_line(p)
        if 'box border' in p:
            colours.box_border = get_colour_from_line(p)
        if 'box text border' in p:
            colours.box_text_border = get_colour_from_line(p)
        if 'box text label' in p:
            colours.box_text_label = get_colour_from_line(p)

        if 'box title text' in p:
            colours.box_title_text = get_colour_from_line(p)
        if 'box text normal' in p:
            colours.box_text = get_colour_from_line(p)
        if 'box sub text' in p:
            colours.box_sub_text = get_colour_from_line(p)
        if 'box input text' in p:
            colours.box_input_text = get_colour_from_line(p)

        if 'box button text highlight' in p:
            colours.box_button_text_highlight = get_colour_from_line(p)
        if 'box button text normal' in p:
            colours.box_button_text = get_colour_from_line(p)
        if 'box button background normal' in p:
            colours.box_button_background = get_colour_from_line(p)
        if 'box button background highlight' in p:
            colours.box_button_background_highlight = get_colour_from_line(p)
        if 'box button border' in p:
            colours.box_check_border = get_colour_from_line(p)

        if "window buttons background" in p:
            colours.window_buttons_bg = get_colour_from_line(p)
        if "window buttons on" in p:
            colours.window_buttons_bg_over = get_colour_from_line(p)
        if "window buttons icon off" in p:
            colours.window_button_icon_off = get_colour_from_line(p)
            colours.window_button_x_off = colours.window_button_icon_off
        if "window buttons icon over" in p:
            colours.window_buttons_icon_over = get_colour_from_line(p)
            colours.window_button_x_on = colours.window_buttons_icon_over
        if "window button x on" in p:
            colours.window_button_x_on = get_colour_from_line(p)
        if "window button x off" in p:
            colours.window_button_x_off = get_colour_from_line(p)
        if "column bar background" in p:
            colours.column_bar_background = get_colour_from_line(p)

        if "artist bio background" in p:
            colours.artist_bio_background = get_colour_from_line(p)
        if "artist bio text" in p:
            colours.artist_bio_text = get_colour_from_line(p)
        # if "panel button off" in p:
        #     colours.corner_button = get_colour_from_line(p)
        # if "panel button on" in p:
        #     colours.corner_button_active = get_colour_from_line(p)

    colours.post_config()
    if colours.lm:
        colours.light_mode()



# Tauon Music Box - Module for DBus interaction

# Copyright © 2015-2019, Taiko2k captain(dot)gxj(at)gmail.com

#     This file is part of Tauon Music Box.
#
#     Tauon Music Box is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     Tauon Music Box is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Lesser General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with Tauon Music Box.  If not, see <http://www.gnu.org/licenses/>.


from gi.repository import GLib, Gdk
import urllib.parse
from t_modules.t_extra import *
import shutil
import os

class Gnome:

    def __init__(self, tauon):

        self.bus_object = None
        self.tauon = tauon
        self.indicator_launched = False
        self.indicator_mode = 0
        self.update_tray_text = None
        self.tray_text = ""

        self.indicator_icon_play = os.path.join(self.tauon.pctl.install_directory, "assets/svg/tray-indicator-play.svg")
        self.indicator_icon_pause = os.path.join(self.tauon.pctl.install_directory, "assets/svg/tray-indicator-pause.svg")
        self.indicator_icon_default = os.path.join(self.tauon.pctl.install_directory, "assets/svg/tray-indicator-default.svg")

        if tauon.prefs.flatpak_mode:

            # This is a workaround to make tray icons visible from outside the sandbox

            export_dir = os.path.join(self.tauon.cache_directory, "icon-export")
            if not os.path.isdir(export_dir):
                os.makedirs(export_dir)

            print("Copy tray icons to data directory...")

            alt = os.path.join(self.tauon.cache_directory, "icon-export/tray-indicator-play.svg")
            if os.path.isfile(self.indicator_icon_play):
                shutil.copy(self.indicator_icon_play, alt)
                self.indicator_icon_play = alt

            alt = os.path.join(self.tauon.cache_directory, "icon-export/tray-indicator-pause.svg")
            if os.path.isfile(self.indicator_icon_pause):
                shutil.copy(self.indicator_icon_pause, alt)
                self.indicator_icon_pause = alt

            alt = os.path.join(self.tauon.cache_directory, "icon-export/tray-indicator-default.svg")
            if os.path.isfile(self.indicator_icon_default):
                shutil.copy(self.indicator_icon_default, alt)
                self.indicator_icon_default = alt

    def focus(self):

        if self.bus_object is not None:
            try:
                # this is what gives us the multi media keys.
                dbus_interface = 'org.gnome.SettingsDaemon.MediaKeys'
                self.bus_object.GrabMediaPlayerKeys("TauonMusicBox", 0, dbus_interface=dbus_interface)
            except:
                # Error connecting to org.gnome.SettingsDaemon.MediaKeys
                pass

    def show_indicator(self):
        if not self.indicator_launched:
            try:
                self.start_indicator()
            except:
                self.tauon.gui.show_message("Failed to start indicator", mode="error")
        else:
            self.indicator.set_status(1)

    def hide_indicator(self):
        if self.indicator_launched:
            self.indicator.set_status(0)

    def indicator_play(self):
        if self.indicator_launched:
            self.indicator.set_icon_full(self.indicator_icon_play, "playing")

    def indicator_pause(self):
        if self.indicator_launched:
            self.indicator.set_icon_full(self.indicator_icon_pause, "paused")

    def indicator_stop(self):
        if self.indicator_launched:
            self.indicator.set_icon_full(self.indicator_icon_default, "default")

    def start_indicator(self):

        pctl = self.tauon.pctl
        tauon = self.tauon

        import gi
        gi.require_version('AppIndicator3', '0.1')
        from gi.repository import Gtk
        from gi.repository import AppIndicator3

        self.indicator = AppIndicator3.Indicator.new("Tauon", self.indicator_icon_default, AppIndicator3.IndicatorCategory.APPLICATION_STATUS)
        self.indicator.set_status(AppIndicator3.IndicatorStatus.ACTIVE)  # 1
        self.indicator.set_title(tauon.t_title)
        self.menu = Gtk.Menu()

        def restore(_):
            tauon.raise_window()

        def menu_quit(_):
            print("Exit via tray")
            tauon.exit()
            self.indicator.set_status(AppIndicator3.IndicatorStatus.PASSIVE)  # 0

        def play_pause(_):
            pctl.play_pause()

        def next(_):
            pctl.advance()

        def back(_):
            pctl.back()

        def update():
            # This is done polling style in a single thread because calling
            # from a different thread seems to cause text to sometimes stall

            while True:

                time.sleep(0.25)
                if tauon.tray_releases <= 0:
                    tauon.tray_lock.acquire()
                tauon.tray_releases -= 1

                if pctl.playing_state in (1, 3):
                    if self.indicator_mode != 1:
                        self.indicator_mode = 1
                        self.indicator_play()
                elif pctl.playing_state == 2:
                    if self.indicator_mode != 2:
                        self.indicator_mode = 2
                        self.indicator_pause()
                else:
                    if self.indicator_mode != 0:
                        self.indicator_mode = 0
                        self.indicator_stop()

                text = ""
                if self.tauon.prefs.tray_show_title:
                    tr = pctl.playing_object()
                    if tr and tr.title and tr.artist:
                        text = tr.artist + " - " + tr.title
                    elif tr and tr.filename:
                        text = tr.filename

                    if pctl.playing_state == 0:
                        text = ""

                if self.indicator_launched:
                    if text != self.tray_text:
                        if text:
                            self.indicator.set_label(" " + text, text)
                            self.indicator.set_title(text)
                        else:
                            self.indicator.set_label("", "")
                            self.indicator.set_title(tauon.t_title)
                        self.tray_text = text

        item = Gtk.MenuItem("Open Tauon Music Box")
        item.connect("activate", restore)
        item.show()
        self.menu.append(item)

        item = Gtk.SeparatorMenuItem()
        item.show()
        self.menu.append(item)

        item = Gtk.MenuItem("Play/Pause")
        item.connect("activate", play_pause)
        item.show()
        self.menu.append(item)

        item = Gtk.MenuItem("Next Track")
        item.connect("activate", next)
        item.show()
        self.menu.append(item)

        item = Gtk.MenuItem("Previous Track")
        item.connect("activate", back)
        item.show()
        self.menu.append(item)

        item = Gtk.SeparatorMenuItem()
        item.show()
        self.menu.append(item)

        item = Gtk.MenuItem("Quit")
        item.connect("activate", menu_quit)
        item.show()
        self.menu.append(item)

        self.menu.show()

        self.indicator.set_menu(self.menu)

        self.indicator.connect("scroll-event", self.scroll)

        self.tauon.gui.tray_active = True
        self.indicator_launched = True

        import threading
        shoot = threading.Thread(target=update)
        shoot.daemon = True
        shoot.start()

    def scroll(self, indicator, steps, direction):
        if direction == Gdk.ScrollDirection.UP:
            self.tauon.pctl.player_volume += 4
            if self.tauon.pctl.player_volume > 100:
                self.tauon.pctl.player_volume = 100
            self.tauon.pctl.set_volume()
        if direction == Gdk.ScrollDirection.DOWN:
            if self.tauon.pctl.player_volume > 4:
                self.tauon.pctl.player_volume -= 4
            else:
                self.tauon.pctl.player_volume = 0
            self.tauon.pctl.set_volume()
        self.tauon.gui.update += 1

    def main(self):

        import dbus
        import dbus.service
        import dbus.mainloop.glib

        prefs = self.tauon.prefs
        gui = self.tauon.gui
        pctl = self.tauon.pctl
        tauon = self.tauon

        if prefs.use_tray:
            self.show_indicator()

        def on_mediakey(comes_from, what):

            if what == 'Play':
                self.tauon.inp.media_key = 'Play'
            elif what == 'Pause':
                self.tauon.inp.media_key = 'Pause'
            elif what == 'Stop':
                self.tauon.inp.media_key = 'Stop'
            elif what == 'Next':
                self.tauon.inp.media_key = 'Next'
            elif what == 'Previous':
                self.tauon.inp.media_key = 'Previous'
            elif what == 'Rewind':
                self.tauon.inp.media_key = 'Rewind'
            elif what == 'FastForward':
                self.tauon.inp.media_key = 'FastForward'
            elif what == 'Repeat':
                self.tauon.inp.media_key = 'Repeat'
            elif what == 'Shuffle':
                self.tauon.inp.media_key = 'Shuffle'

            if self.tauon.inp.media_key:
                gui.update = 1

        try:
            # set up the glib main loop.
            dbus.mainloop.glib.DBusGMainLoop(set_as_default=True)

            bus = dbus.Bus(dbus.Bus.TYPE_SYSTEM)
            bus_object = bus.get_object('org.freedesktop.login1',
                                        '/org/freedesktop/login1')

            iface = dbus.Interface(bus_object,
                                   dbus_interface='org.freedesktop.login1.Manager')

            tauon.sleep_lock = iface.Inhibit("sleep", "Tauon Music Box", "Pause music on sleep", "delay")
            tauon.shutdown_lock = iface.Inhibit("shutdown", "Tauon Music Box", "Save data to disk", "delay")

            def update_play_lock():
                if prefs.block_suspend:
                    if pctl.playing_state in (1, 3) and tauon.play_lock is None:
                        tauon.play_lock = iface.Inhibit("sleep", "Tauon Music Box", "Audio is playing", "block")
                    elif pctl.playing_state not in (1, 3) and tauon.play_lock is not None:
                        del tauon.play_lock
                        tauon.play_lock = None
                elif tauon.play_lock is not None:
                    del tauon.play_lock
                    tauon.play_lock = None

            tauon.update_play_lock = update_play_lock

            def PrepareForSleep(active):

                if active == 1 and tauon.sleep_lock is not None:
                    print("System is suspending!")
                    if pctl.playing_state in (1, 3):
                        tauon.pctl.pause()
                    del tauon.sleep_lock
                    tauon.sleep_lock = None

                elif active == 0 and tauon.sleep_lock is None:
                    tauon.sleep_lock = iface.Inhibit("sleep", "Tauon Music Box", "Pause music on sleep", "delay")

            def PrepareForShutdown(active):
                print("The system is shutting down!")
                tauon.quick_close = True
                tauon.exit()

            iface.connect_to_signal("PrepareForSleep", PrepareForSleep)
            iface.connect_to_signal("PrepareForShutdown", PrepareForShutdown)


        except:
            print("Failure to connect to login1")

        if prefs.mkey:
            try:
                bus = dbus.Bus(dbus.Bus.TYPE_SESSION)
                bus_object = bus.get_object('org.gnome.SettingsDaemon.MediaKeys',
                                            '/org/gnome/SettingsDaemon/MediaKeys')

                self.bus_object = bus_object

                # this is what gives us the multi media keys.
                dbus_interface = 'org.gnome.SettingsDaemon.MediaKeys'
                bus_object.GrabMediaPlayerKeys("TauonMusicBox", 0,
                                               dbus_interface=dbus_interface)

                # connect_to_signal registers our callback function.
                bus_object.connect_to_signal('MediaPlayerKeyPressed',
                                             on_mediakey)
            except:
                print("Could not connect to gnome media keys")

        # ----------

        # t_bus = dbus.Bus(dbus.Bus.TYPE_SESSION)
        # t_bus_name = dbus.service.BusName('com.github.taiko2k.tauonmb', t_bus)  # This object must be kept alive
        #
        # class T(dbus.service.Object):
        #     @dbus.service.method("com.github.taiko2k.tauonmb",
        #                          in_signature='a{sv}', out_signature='')
        #     def start(self, options={}):
        #         print("START")
        #
        #     def __init__(self, object_path):
        #         dbus.service.Object.__init__(self, t_bus, object_path, bus_name=t_bus_name)
        #
        # pctl.sgl = T("/")

        # ----------
        if prefs.enable_mpris:
            try:
                bus = dbus.Bus(dbus.Bus.TYPE_SESSION)
                bus_name = dbus.service.BusName('org.mpris.MediaPlayer2.tauon', bus)  # This object must be kept alive

                class MPRIS(dbus.service.Object):

                    def update(self, force=False):

                        changed = {}

                        if pctl.playing_state == 1 or pctl.playing_state == 3:
                            if self.player_properties['PlaybackStatus'] != 'Playing':
                                self.player_properties['PlaybackStatus'] = 'Playing'
                                changed['PlaybackStatus'] = self.player_properties['PlaybackStatus']
                        elif pctl.playing_state == 0:
                            if self.player_properties['PlaybackStatus'] != 'Stopped':
                                self.player_properties['PlaybackStatus'] = 'Stopped'
                                changed['PlaybackStatus'] = self.player_properties['PlaybackStatus']
                        elif pctl.playing_state == 2:
                            if self.player_properties['PlaybackStatus'] != 'Paused':
                                self.player_properties['PlaybackStatus'] = 'Paused'
                                changed['PlaybackStatus'] = self.player_properties['PlaybackStatus']

                        if pctl.player_volume / 100 != self.player_properties['Volume']:
                            self.player_properties['Volume'] = pctl.player_volume / 100
                            changed['Volume'] = self.player_properties['Volume']

                        if pctl.playing_object() is not None and (pctl.playing_object().index != self.playing_index or force):
                            track = pctl.playing_object()
                            self.playing_index = track.index
                            id = f"/com/tauon/{track.index}/{abs(pctl.playlist_playing_position)}"
                            if pctl.playing_state == 3:
                                id = "/com/tauon/radio"

                            d = {
                                'mpris:trackid': dbus.ObjectPath(id),
                                'mpris:length': dbus.Int64(int(pctl.playing_length * 1000000)),
                                'xesam:album': track.album,
                                'xesam:albumArtist': dbus.Array([track.album_artist]),
                                'xesam:artist': dbus.Array([track.artist]),
                                'xesam:title': track.title,
                                'xesam:asText': track.lyrics,
                                'xesam:autoRating': star_count2(tauon.star_store.get(track.index)),
                                'xesam:composer': dbus.Array([track.composer]),
                                'tauon:loved': tauon.love(False, track.index),
                                # added by msmafra
                                'xesam:comment': dbus.Array([track.comment]),
                                'xesam:genre': dbus.Array([track.genre])

                            }
                            try:
                                d['xesam:url'] = "file://" + urllib.parse.quote(track.fullpath)
                            except:
                                print("Uri encode error")

                            try:
                                i_path = tauon.thumb_tracks.path(track)
                                if i_path is not None:
                                    d['mpris:artUrl'] = 'file://' + urllib.parse.quote(i_path)
                            except Exception as e:
                                print(str(e))
                                print("Thumbnail error")
                                print(track.fullpath)

                            self.update_progress()

                            self.player_properties['Metadata'] = dbus.Dictionary(d, signature='sv')
                            changed['Metadata'] = self.player_properties['Metadata']

                            if pctl.playing_state == 3 and self.player_properties['CanPause'] is True:
                                self.player_properties['CanPause'] = False
                                self.player_properties['CanSeek'] = False
                                changed['CanPause'] = self.player_properties['CanPause']
                                changed['CanSeek'] = self.player_properties['CanSeek']
                            elif pctl.playing_state == 1 and self.player_properties['CanPause'] is False:
                                self.player_properties['CanPause'] = True
                                self.player_properties['CanSeek'] = True
                                changed['CanPause'] = self.player_properties['CanPause']
                                changed['CanSeek'] = self.player_properties['CanSeek']

                        if len(changed) > 0:
                            try:
                                self.PropertiesChanged('org.mpris.MediaPlayer2.Player', changed, [])
                            except Exception as e:
                                print("Error updating MPRIS")
                                print(str(e))
                                print(changed)
                                print(pctl.playing_object().fullpath)

                    def update_progress(self):
                        self.player_properties['Position'] = dbus.Int64(int(pctl.playing_time * 1000000))

                    def update_shuffle(self):
                        self.player_properties['Shuffle'] = pctl.random_mode
                        self.PropertiesChanged('org.mpris.MediaPlayer2.Player', {"Shuffle": pctl.random_mode}, [])

                    def update_loop(self):
                        self.player_properties['LoopStatus'] = self.get_loop_status()
                        self.PropertiesChanged('org.mpris.MediaPlayer2.Player', {"LoopStatus": self.get_loop_status()}, [])

                    def __init__(self, object_path):
                        # dbus.service.Object.__init__(self, bus_name, object_path)
                        dbus.service.Object.__init__(self, bus, object_path, bus_name=bus_name)

                        self.playing_index = -1

                        self.root_properties = {
                            'CanQuit': True,
                            #'Fullscreen'
                            #'CanSetFullscreen'
                            'CanRaise': True,
                            'HasTrackList': False,
                            'Identity': tauon.t_title,
                            'DesktopEntry': tauon.t_id,
                            'SupportedUriSchemes': dbus.Array([dbus.String("file")]),
                            'SupportedMimeTypes': dbus.Array([
                                 dbus.String("audio/mpeg"),
                                 dbus.String("audio/flac"),
                                 dbus.String("audio/ogg"),
                                 dbus.String("audio/m4a"),
                                 ])
                        }

                        self.player_properties = {
                            'PlaybackStatus': 'Stopped',
                            'LoopStatus': self.get_loop_status(),
                            'Rate': 1.0,
                            'Shuffle': pctl.random_mode,
                            'Volume': pctl.player_volume / 100,
                            'Position': dbus.Int64(0),
                            'MinimumRate': 1.0,
                            'MaximumRate': 1.0,
                            'CanGoNext': True,
                            'CanGoPrevious': True,
                            'CanPlay': True,
                            'CanPause': True,
                            'CanSeek': True,
                            'CanControl': True,
                            'Metadata': dbus.Dictionary({}, signature='sv')
                        }

                    def get_loop_status(self):
                        if pctl.repeat_mode:
                            if pctl.album_repeat_mode:
                                return "Playlist"
                            return "Track"
                        return "None"

                    @dbus.service.method(dbus_interface='org.mpris.MediaPlayer2')
                    def Raise(self):
                        gui.request_raise = True
                        tauon.wake()

                    @dbus.service.method(dbus_interface='org.mpris.MediaPlayer2')
                    def Quit(self):
                        tauon.wake()
                        tauon.exit()

                    @dbus.service.method(dbus_interface=dbus.PROPERTIES_IFACE,
                                    in_signature='ss', out_signature='v')
                    def Get(self, interface_name, property_name):
                        if interface_name == 'org.mpris.MediaPlayer2':
                            #return self.GetAll(interface_name)[property_name]
                            return self.root_properties[property_name]
                        elif interface_name == 'org.mpris.MediaPlayer2.Player':
                            return self.player_properties[property_name]

                    @dbus.service.method(dbus_interface=dbus.PROPERTIES_IFACE,
                                    in_signature='s', out_signature='a{sv}')
                    def GetAll(self, interface_name):

                        if interface_name == 'org.mpris.MediaPlayer2':
                            return self.root_properties
                        elif interface_name == 'org.mpris.MediaPlayer2.Player':
                            return self.player_properties
                        else:
                            return {}

                    @dbus.service.method(dbus_interface=dbus.PROPERTIES_IFACE,
                                    in_signature='ssv', out_signature='')
                    def Set(self, interface_name, property_name, value):
                        if interface_name == 'org.mpris.MediaPlayer2.Player':
                            tauon.wake()
                            if property_name == "Volume":
                                pctl.player_volume = min(max(int(value * 100), 0), 100)
                                pctl.set_volume()
                                gui.update += 1
                            if property_name == "Shuffle":
                                pctl.random_mode = bool(value)
                                self.update_shuffle()
                                gui.update += 1
                            if property_name == "LoopStatus":
                                if value == "None":
                                    tauon.menu_repeat_off()
                                elif value == "Track":
                                    tauon.menu_set_repeat()
                                elif value == "Playlist":
                                    tauon.menu_album_repeat()
                                gui.update += 1

                        if interface_name == 'org.mpris.MediaPlayer2':
                            pass

                    @dbus.service.signal(dbus_interface=dbus.PROPERTIES_IFACE,
                                    signature='sa{sv}as')
                    def PropertiesChanged(self, interface_name, change, inval):
                        pass

                    @dbus.service.method(dbus_interface='org.mpris.MediaPlayer2.Player')
                    def Next(self):
                        tauon.wake()
                        pctl.advance()
                        pass

                    @dbus.service.method(dbus_interface='org.mpris.MediaPlayer2.Player')
                    def Previous(self):
                        tauon.wake()
                        pctl.back()
                        pass

                    @dbus.service.method(dbus_interface='org.mpris.MediaPlayer2.Player')
                    def Pause(self):
                        pctl.pause_only()

                    @dbus.service.method(dbus_interface='org.mpris.MediaPlayer2.Player')
                    def PlayPause(self):
                        tauon.wake()
                        if pctl.playing_state == 3:
                            pctl.stop()  # Stop if playing radio
                        else:
                            pctl.play_pause()

                    @dbus.service.method(dbus_interface='org.mpris.MediaPlayer2.Player')
                    def Stop(self):
                        pctl.stop()

                    @dbus.service.method(dbus_interface='org.mpris.MediaPlayer2.Player')
                    def Play(self):
                        tauon.wake()
                        pctl.play()

                    @dbus.service.method(dbus_interface='org.mpris.MediaPlayer2.Player')
                    def Seek(self, offset):
                        pctl.seek_time(pctl.playing_time + (offset / 1000000))

                    @dbus.service.method(dbus_interface='org.mpris.MediaPlayer2.Player')
                    def SetPosition(self, id, position):
                        pctl.seek_time(position / 1000000)

                        self.player_properties['Position'] = dbus.Int64(int(position))
                        self.Seeked(pctl.playing_time)

                    @dbus.service.method(dbus_interface='org.mpris.MediaPlayer2.Player')
                    def OpenUri(self, uri):
                        tauon.wake()
                        tauon.open_uri(uri)

                    @dbus.service.method(dbus_interface='org.mpris.MediaPlayer2.Player')
                    def LovePlaying(self):
                        if not tauon.love(set=False):
                            tauon.love(set=True, no_delay=True)
                            self.update(True)
                            gui.pl_update += 1

                    @dbus.service.method(dbus_interface='org.mpris.MediaPlayer2.Player')
                    def UnLovePlaying(self):
                        if tauon.love(set=False):
                            tauon.love(set=True, no_delay=True)
                            self.update(True)
                            gui.pl_update += 1

                    @dbus.service.signal(dbus_interface='org.mpris.MediaPlayer2.Player')
                    def Seeked(self, position):
                        pass

                    def seek_do(self, seconds):
                        self.Seeked(dbus.Int64(int(seconds * 1000000)))

                pctl.mpris = MPRIS("/org/mpris/MediaPlayer2")

            except:
                print("MPRIS2 CONNECT FAILED")

        mainloop = GLib.MainLoop()
        mainloop.run()



# Tauon Music Box - OGG Broadcasting module

# Copyright © 2020, Taiko2k captain(dot)gxj(at)gmail.com

#     This file is part of Tauon Music Box.
#
#     Tauon Music Box is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     Tauon Music Box is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Lesser General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with Tauon Music Box.  If not, see <http://www.gnu.org/licenses/>.


import subprocess
import time
import io
import shlex
import fcntl, os
from t_modules.t_extra import Timer

def enc(tauon):

    pctl = tauon.pctl
    prefs = tauon.prefs

    class Enc:
        def __init__(self):
            self.encoder = None
            self.decoder = None

            self.raw_buffer = io.BytesIO()
            self.raw_buffer_size = 0

            self.output_buffer = io.BytesIO()
            self.output_buffer_size = 0

            self.temp_buffer = io.BytesIO()
            self.temp_buffer_size = 0

            self.stream_time = Timer()
            self.bytes_sent = 0

            self.track_bytes_sent = 0

            self.dry = 0

        def get_decode_command(self, target, start):
            s = start
            s, ms = divmod(s, 1)
            m, ss = divmod(s, 60)
            hh, mm = divmod(m, 60)
            ms *= 10
            t = f"{str(int(hh)).zfill(2)}:{str(int(mm)).zfill(2)}:{str(int(ss)).zfill(2)}.{str(round(ms))}"

            return ['ffmpeg', "-loglevel", "quiet", "-i", target, "-ss", t, "-acodec", "pcm_s16le", "-f", "s16le", "-ac", "2", "-ar",  # -re
                    "48000", "-"]

        def main(self):

            while True:
                if not pctl.broadcast_active and not pctl.broadcastCommandReady:
                    return
                if pctl.broadcastCommandReady:
                    command = pctl.broadcastCommand
                    pctl.playerCommand = ""
                    pctl.broadcastCommandReady = False

                    if command == "encstop":
                        # print("Stopping broadcast...")
                        pctl.broadcast_active = False
                        time.sleep(1)
                        self.decoder.terminate()
                        self.encoder.terminate()
                        self.encoder = None
                        self.decoder = None
                        self.track_bytes_sent = 0
                        self.output_buffer_size = 0
                        self.raw_buffer = io.BytesIO()
                        self.output_buffer = io.BytesIO()
                        self.temp_buffer = io.BytesIO()
                        self.dry = 0
                        self.bytes_sent = 0
                        self.stream_time.set()
                        tauon.chunker.chunks.clear()
                        tauon.chunker.headers.clear()
                        tauon.chunker.master_count = 0
                        print("Broadcast stopped")

                        pctl.broadcast_time = 0

                    if command == "encstart":
                        # print("Start broadcast...")
                        target = pctl.target_open
                        # print(f"URI = {target}")
                        pctl.broadcast_active = True
                        # print("Start encoder")
                        #cmd = shlex.split("opusenc --raw --raw-rate 48000 - -")
                        cmd = ["ffmpeg", "-loglevel", "quiet", "-f", "s16le", "-ar", "48000", "-ac", "2", "-i", "pipe:0", '-f', "opus", "-c:a", "libopus", "pipe:1"]
                        # cmd = shlex.split("oggenc --raw --raw-rate 48000 -")
                        self.encoder = subprocess.Popen(cmd, stdin=subprocess.PIPE, stdout=subprocess.PIPE)
                        fcntl.fcntl(self.encoder.stdout.fileno(), fcntl.F_SETFL, os.O_NONBLOCK)

                        # print("Begin decode of file")
                        cmd = self.get_decode_command(target, pctl.b_start_time)
                        # print(cmd)
                        self.decoder = subprocess.Popen(cmd, stdin=subprocess.PIPE, stdout=subprocess.PIPE)
                        time.sleep(0.1)
                        self.stream_time.force_set(6)
                        print("Broadcast started")

                    if command == "cast-next":
                        target = pctl.target_open
                        # print(f"URI = {target}")

                        self.decoder.terminate()
                        cmd = self.get_decode_command(target, 0)
                        self.decoder = subprocess.Popen(cmd, stdin=subprocess.PIPE, stdout=subprocess.PIPE)
                        time.sleep(0.1)
                        # print("started next")
                        self.track_bytes_sent = 0

                    if command == "encseek":
                        target = pctl.target_open
                        start = pctl.b_start_time + pctl.broadcast_seek_position
                        # print(f"URI = {target}")
                        self.decoder.terminate()
                        cmd = self.get_decode_command(target, start)
                        self.decoder = subprocess.Popen(cmd, stdin=subprocess.PIPE, stdout=subprocess.PIPE)
                        time.sleep(0.1)
                        # print("started next")
                        self.track_bytes_sent = pctl.broadcast_seek_position * (48000 * (16 / 8) * 2)

                if self.decoder:

                    pctl.broadcast_time = self.track_bytes_sent / (48000 * (16 / 8) * 2)

                    st = self.stream_time.get()
                    ss = self.bytes_sent / (48000 * (16 / 8) * 2)
                    #print((st, ss))

                    if ss < st:
                        # We owe:
                        owed_seconds = st - ss
                        owed_bytes = owed_seconds * (48000 * (16 / 8) * 2)

                        while owed_bytes > 0:
                            #print("PUMP")
                            # Pump data out of decoder
                            data = self.decoder.stdout.read(int(48000 * (16 / 8) * 2))
                            #print(data)
                            if not data:
                                self.dry += 1

                                if owed_seconds > 0.1 and self.dry > 2:
                                    #print("SILENCE")
                                    data = b"\x00" * 19200
                                else:
                                    break
                            else:
                                self.dry = 0
                            self.raw_buffer.write(data)
                            self.raw_buffer_size += len(data)
                            self.bytes_sent += len(data)
                            self.track_bytes_sent += len(data)
                            owed_bytes -= len(data)
                            break

                    if not data:
                        #print("No more decoded data...")
                        time.sleep(0.01)

                    # Push data into encoder
                    if self.raw_buffer_size > 0:
                        self.raw_buffer.seek(0)
                        data = self.raw_buffer.read(self.raw_buffer_size)
                        self.encoder.stdin.write(data)

                        # Reset the buffer
                        self.raw_buffer_size = 0
                        self.raw_buffer.seek(0)

                    # Receive encoded data
                    data = self.encoder.stdout.read()
                    if data:
                        #print("WRITE")
                        self.output_buffer.write(data)
                        self.output_buffer_size += len(data)

                    # Split OGG pages
                    if self.output_buffer_size > 12000:

                        self.output_buffer.seek(6)
                        gp = self.output_buffer.read(8)
                        gp = int.from_bytes(gp, 'big')


                        self.output_buffer.seek(26)
                        cont = self.output_buffer.read(1)
                        cont = int.from_bytes(cont, 'big')
                        #print(f"{cont} segments")
                        total = cont
                        while cont:
                            value = self.output_buffer.read(1)
                            value = int.from_bytes(value, 'big')
                            #print(f"value {value}")
                            cont -= 1
                            total += value

                        total = total + 27

                        self.output_buffer.seek(0, 2)
                        # self.output_buffer.seek(0)
                        # print(self.output_buffer.read(4))
                        # self.output_buffer.seek(0, 2)

                        if self.output_buffer_size >= total:
                            # Extract the first complete page
                            self.output_buffer.seek(0)
                            page = self.output_buffer.read(total)

                            # Save the page
                            if gp == 0:
                                tauon.chunker.headers.append(page)
                            else:
                                tauon.chunker.chunks[tauon.chunker.master_count] = page
                            tauon.chunker.master_count += 1
                            d = tauon.chunker.master_count - 30
                            if d > 1:
                                del tauon.chunker.chunks[d]

                            # print(f"Received page {tauon.chunker.master_count}")

                            # Reset the buffer with the remainder
                            self.temp_buffer.seek(0)
                            self.temp_buffer.write(self.output_buffer.read())
                            self.temp_buffer.seek(0)
                            del self.output_buffer
                            self.output_buffer = io.BytesIO()
                            self.output_buffer.seek(0)
                            self.output_buffer.write(self.temp_buffer.read())
                            self.output_buffer_size = self.output_buffer.tell()
                            del self.temp_buffer
                            self.temp_buffer = io.BytesIO()

    en = Enc()
    en.main()



# Copyright © 2019-2020, Taiko2k captain(dot)gxj(at)gmail.com

#     This file is part of Tauon Music Box.
#
#     Tauon Music Box is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     Tauon Music Box is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Lesser General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with Tauon Music Box.  If not, see <http://www.gnu.org/licenses/>.


from gi import require_version
require_version('Rsvg', '2.0')
from gi.repository import Rsvg
import cairo
import os


def render_icons(source_directory, output_directory, scale):

    targets = []
    # Verify svg files exist
    for file in os.listdir(source_directory):
        if file.endswith(".svg") and os.path.isfile(os.path.join(source_directory, file)):
            targets.append(file)

    if not os.path.exists(output_directory):
        os.makedirs(output_directory)

    # Render svg files to png
    for file in targets:
        name = os.path.splitext(file)[0]
        in_path = os.path.join(source_directory, file)
        out_path = os.path.join(output_directory, name + ".png")

        handle = Rsvg.Handle()
        svg = handle.new_from_file(in_path)

        unscaled_width = svg.props.width
        unscaled_height = svg.props.height

        width = unscaled_width * scale
        height = unscaled_height * scale

        svg_surface = cairo.SVGSurface(None, width, height)
        svg_context = cairo.Context(svg_surface)
        svg_context.save()
        svg_context.scale(width/unscaled_width, height/unscaled_height)
        svg.render_cairo(svg_context)
        svg_context.restore()

        svg_surface.write_to_png(out_path)



# Tauon Music Box - Configuration file module

# Copyright © 2015-2019, Taiko2k captain(dot)gxj(at)gmail.com

#     This file is part of Tauon Music Box.
#
#     Tauon Music Box is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     Tauon Music Box is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Lesser General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with Tauon Music Box.  If not, see <http://www.gnu.org/licenses/>.



# The purpose of this module is to update, parse and write to a configuration file

import os


class Config:

    def __init__(self):

        self.live = []
        self.old = []

    def reset(self):
        self.live.clear()
        self.old.clear()

    def add_text(self, text):

        self.live.append(['comment', text])

    def add_comment(self, text):

        self.live.append(['comment', "# " + text])

    def br(self):

        self.live.append(['comment', ""])

    def update_value(self, key, value):

        for item in self.live:
            if item[0] != "comment":
                if item[1] == key:
                    item[2] = value
                    break

    def load(self, path):

        if os.path.isfile(path):
            with open(path, encoding="utf_8") as f:
                self.old = f.readlines()

    def dump(self, path):

        # if os.path.exists(path) and not os.access("test.conf", os.W_OK):
        #     print("ERROR! Config file cannot be written")
        #     return

        with open(path, 'w', encoding="utf_8") as f:

            for item in self.live:

                if item[0] == 'comment':
                    f.write(item[1])
                    f.write(os.linesep)
                    continue

                f.write(item[1])
                f.write(" = ")

                if item[0] == 'bool':
                    if item[2] is True:
                        f.write("true")
                    else:
                        f.write("false")

                if item[0] == 'string':
                    f.write('"')
                    f.write(item[2])
                    f.write('"')

                if item[0] == 'int':
                    f.write(str(item[2]))

                if item[0] == 'float':
                    f.write(str(item[2]))

                if item[3]:
                    f.write("  # ")
                    f.write(item[3])

                f.write(os.linesep)

            f.write(os.linesep)

    def sync_add(self, type, key, default_value, comment=""):

        got_old = False
        old_value = None

        for row in self.old:
            row = row.split(" #", 1)[0]
            if "=" in row:
                if row.split("=", 1)[0].strip() == key:
                    old_value = row.split("=", 1)[1].strip()
                    if old_value:
                        got_old = True
                        break

        if type == 'bool':
            if got_old:
                if old_value == 'true':
                    self.live.append(['bool', key, True, comment])
                    return True
                if old_value == 'false':
                    self.live.append(['bool', key, False, comment])
                    return False
            self.live.append(['bool', key, default_value, comment])
            return default_value

        if type == 'string':

            if old_value is None:
                self.live.append(['string', key, default_value, comment])
                return default_value

            #old_value = old_value.strip('"')
            if old_value and old_value[0] == old_value[-1] == "\"":
                old_value = old_value[1:-1]

            if not got_old:
                self.live.append(['string', key, default_value, comment])
                return default_value

            self.live.append(['string', key, old_value, comment])
            return old_value

            # if got_old:
            #     old_value = old_value.strip('"')
            #     if old_value:
            #         self.live.append(['string', key, old_value, comment])
            #         return old_value
            #
            # self.live.append(['string', key, default_value, comment])
            # return default_value

        if type == 'int':
            if got_old and old_value.isdigit():
                old_value = int(old_value)
                self.live.append(['int', key, old_value, comment])
                return old_value

            self.live.append(['int', key, default_value, comment])
            return default_value

        if type == 'float':
            if got_old:
                try:
                    old_value = float(old_value)
                    self.live.append(['float', key, old_value, comment])
                    return old_value
                except:
                    print("Warning: Config file contains invalid float")
                    pass
            self.live.append(['float', key, default_value, comment])
            return default_value


